import copy
import unittest
import json
from aminer.parsing.JsonStringModelElement import JsonStringModelElement, JsonAccessObject
from aminer.parsing.MatchContext import MatchContext
from aminer.parsing.MatchElement import MatchElement
from aminer.parsing.DecimalFloatValueModelElement import DecimalFloatValueModelElement
from unit.TestBase import TestBase, DummyMatchContext, DummyFixedDataModelElement, DummyFirstMatchModelElement


class JsonStringModelElementTest(TestBase):
    """Unittests for the JsonStringModelElement."""
    id_ = "json"
    path = "path"
    strict = False
    ignore_null = True

    def test1get_id(self):
        """Test if get_id works properly."""
        host = DummyFixedDataModelElement("host", b"www.google.com")
        user = DummyFixedDataModelElement("user", b"foobar")

        key_parser_dict = { "host": host, "user": user }

        json_me = JsonStringModelElement(self.id_, key_parser_dict, self.strict, self.ignore_null)
        self.assertEqual(json_me.get_id(), self.id_)

    def test2get_match_element_valid_match(self):
        """Parses a json-file and compares if the configured ModelElements are parsed properly"""
        host = DummyFixedDataModelElement("host", b"www.google.com")
        user = DummyFixedDataModelElement("user", b"foobar")

        key_parser_dict = { "host": host, "user": user }

        json_model_element = JsonStringModelElement(self.id_, key_parser_dict, self.strict, self.ignore_null)
        data = b'{"host": "www.google.com", "user": "foobar", "one": "two"}'
        match_context = DummyMatchContext(data)
        match_element = json_model_element.get_match_element(self.path, match_context)
        self.assertEqual(2, len(match_element.children))
        self.assertEqual(b"www.google.com", match_element.children[0].get_match_object())
        self.assertEqual(b"foobar", match_element.children[1].get_match_object())

    def test3strict_mode(self):
        """Parses a json-file and compares if the configured ModelElements are parsed properly with strict_mode"""
        host = DummyFixedDataModelElement("host", b"www.google.com")
        user = DummyFixedDataModelElement("user", b"foobar")
        path = DummyFixedDataModelElement("path", b"/index.html")

        key_parser_dict = { "host": { "server": host }, "user": user }

        # Sets strict_mode to True
        json_model_element = JsonStringModelElement(self.id_, key_parser_dict, True, self.ignore_null)
        # "one": "two" is too much
        data = b'{"host": {"server": "www.google.com"}, "user": "foobar", "one": "two"}'
        match_context = DummyMatchContext(data)
        match_element = json_model_element.get_match_element(self.path, match_context)
        self.assertEqual(None,match_element)

        # Sets one more element
        key_parser_dict = { "host": { "server": host }, "user": user, "path": path }

        # Sets strict_mode to True
        json_model_element = JsonStringModelElement(self.id_, key_parser_dict, True)
        # "one": "two" is too much
        data = b'{"host": {"server": "www.google.com"}, "user": "foobar", "one": "two"}'
        match_context = DummyMatchContext(data)
        match_element = json_model_element.get_match_element(self.path, match_context)
        self.assertEqual(None,match_element)

        # Sets the logdata to the exact configuration-json.
        data = b'{"host": {"server": "www.google.com"}, "user": "foobar", "path": "/index.html"}'
        match_context = DummyMatchContext(data)
        match_element = json_model_element.get_match_element(self.path, match_context)
        self.assertEqual(3,len(match_element.children))
        self.assertEqual(b"www.google.com", match_element.children[0].get_match_object())
        self.assertEqual(b"foobar", match_element.children[1].get_match_object())
        self.assertEqual(b"/index.html", match_element.children[2].get_match_object())

    def test4ignore_null(self):
        """Parses a json-file with ignore_null and compares if the configured ModelElements are parsed properly"""
        host = DummyFixedDataModelElement("host", b"www.google.com")
        user = DummyFixedDataModelElement("user", b"foobar")

        key_parser_dict = { "host": host, "user": user }

        # Set ignore_null to True and strict to False
        json_model_element = JsonStringModelElement(self.id_, key_parser_dict, False, True)
        # Set user to null
        data = b'{"host": "www.google.com", "user": null, "one": "two"}'
        match_context = DummyMatchContext(data)
        match_element = json_model_element.get_match_element(self.path, match_context)
        # Line must be parsed but without "user":
        self.assertEqual(1, len(match_element.children))
        self.assertEqual(b"www.google.com", match_element.children[0].get_match_object())
        
        # set ignore_null to False and strict to False
        json_model_element = JsonStringModelElement(self.id_, key_parser_dict, False, False)
        # Set user to null
        data = b'{"host": "www.google.com", "user": null, "one": "two"}'
        match_context = DummyMatchContext(data)
        match_element = json_model_element.get_match_element(self.path, match_context)
        # expect an unparsed line
        self.assertEqual(None,match_element)

        # set example user to empty string
        user = DummyFixedDataModelElement("user", b"")
        key_parser_dict2 = { "host": host, "user": user }

        # Set ignore_null to False in order to pass b"" to the subparser. Strict is False
        json_model_element = JsonStringModelElement(self.id_, key_parser_dict2, False, False)
        # Set user to null
        data = b'{"host": "www.google.com", "user": null}'
        match_context = DummyMatchContext(data)
        match_element = json_model_element.get_match_element(self.path, match_context)
        # Line must be parsed:
        self.assertEqual(2, len(match_element.children))
        self.assertEqual(b"www.google.com", match_element.children[0].get_match_object())
        self.assertEqual(b"", match_element.children[1].get_match_object())

        # Set ignore_null to True and strict to True
        json_model_element = JsonStringModelElement(self.id_, key_parser_dict, True, True)
        # Set user to null
        data = b'{"host": "www.google.com", "user": null}'
        match_context = DummyMatchContext(data)
        match_element = json_model_element.get_match_element(self.path, match_context)
        # Line must be parsed but without "user":
        self.assertEqual(1, len(match_element.children))
        self.assertEqual(b"www.google.com", match_element.children[0].get_match_object())
        
        # set ignore_null to False and strict to True
        json_model_element = JsonStringModelElement(self.id_, key_parser_dict, True, False)
        # Set user to null
        data = b'{"host": "www.google.com", "user": null}'
        match_context = DummyMatchContext(data)
        match_element = json_model_element.get_match_element(self.path, match_context)
        # expect an unparsed line
        self.assertEqual(None,match_element)

        # set example user to empty string
        user = DummyFixedDataModelElement("user", b"")
        key_parser_dict2 = { "host": host, "user": user }

        # Set ignore_null to False in order to pass b"" to the subparser. Strict is True
        json_model_element = JsonStringModelElement(self.id_, key_parser_dict2, True, False)
        # Set user to null
        data = b'{"host": "www.google.com", "user": null}'
        match_context = DummyMatchContext(data)
        match_element = json_model_element.get_match_element(self.path, match_context)
        # Line must be parsed:
        self.assertEqual(2, len(match_element.children))
        self.assertEqual(b"www.google.com", match_element.children[0].get_match_object())
        self.assertEqual(b"", match_element.children[1].get_match_object())


class JsonAccessObjectTest(TestBase):
    
    def test1get_id(self):
        """Parses a dictionary and see if everything is flattened properly"""
        d = {'a': 'b', 'c': {'w': 'g', 'rata': 'mahatta', 'tic': {'tac': 'toe'}, 'brat': ['worst','wuast',{'key': ['wurst','fleisch'], 'food': 'veggie'},'blues'],'bist': 'narrisch'}, 'foo': 'bar'}
        """
        a: b
        c.w: g
        c.rata: mahatta
        c.tic.tac: toe
        c.brat[0]: worst
        c.brat[1]: wuast
        c.brat[2].key[0]: wurst
        c.brat[2].key[1]: fleisch
        c.brat[2].food: veggie
        c.brat[3]: blues
        foo: bar
        """
        jao =  JsonAccessObject(d)
        self.assertTrue(jao.collection['a'])
        self.assertTrue(jao.collection['c.w'])
        self.assertTrue(jao.collection['c.rata'])
        self.assertTrue(jao.collection['c.tic.tac'])
        self.assertTrue(jao.collection['c.brat[0]'])
        self.assertTrue(jao.collection['c.brat[1]'])
        self.assertTrue(jao.collection['c.brat[2].key[0]'])
        self.assertTrue(jao.collection['c.brat[2].key[1]'])
        self.assertTrue(jao.collection['c.brat[2].food'])
        self.assertTrue(jao.collection['c.brat[3]'])
        self.assertTrue(jao.collection['c.bist'])
        self.assertTrue(jao.collection['foo'])
        self.assertEqual(12,len(jao.collection))
        self.assertEqual("b",jao.collection["a"]["value"])
        self.assertEqual("g",jao.collection["c.w"]["value"])
        self.assertEqual("mahatta",jao.collection["c.rata"]["value"])
        self.assertEqual("toe",jao.collection["c.tic.tac"]["value"])
        self.assertEqual("worst",jao.collection["c.brat[0]"]["value"])
        self.assertEqual("wuast",jao.collection["c.brat[1]"]["value"])
        self.assertEqual("wurst",jao.collection["c.brat[2].key[0]"]["value"])
        self.assertEqual("fleisch",jao.collection["c.brat[2].key[1]"]["value"])
        self.assertEqual("veggie",jao.collection["c.brat[2].food"]["value"])
        self.assertEqual("blues",jao.collection["c.brat[3]"]["value"])
        self.assertEqual("bar",jao.collection["foo"]["value"])




import unittest
from aminer.analysis.HistogramAnalysis import LinearNumericBinDefinition, ModuloTimeBinDefinition, HistogramData, HistogramAnalysis, \
    PathDependentHistogramAnalysis
from aminer.input.LogAtom import LogAtom
from aminer.parsing.ParserMatch import ParserMatch
import time
from datetime import datetime
from aminer.parsing.MatchElement import MatchElement
from unit.TestBase import TestBase


class HistogramAnalysisTest(TestBase):
    """Unittests for the HistogramAnalysis."""

    __expected_string_histogram_analysis = '%s Histogram report\n%s: "%s" (%d lines)\n  Histogram report from %s till %s\n  %s\n\n'
    __expected_string_path_dependent_histogram_analysis = '%s Histogram report\n%s: "%s" (%d lines)\n  Path histogram report from %s ' \
                                                          'till %s\n%s\n\n'

    match_crontab = 'match/crontab'
    datetime_format_string = '%Y-%m-%d %H:%M:%S'

    def test1linear_numeric_bin_definition_get_bin_names(self):
        """This test case aims to test the functionality of the LinearNumericBinDefinition's get_bin_names method."""
        linear_numeric_bin_definition = LinearNumericBinDefinition(0, 1, 10, True)
        self.assertEqual(linear_numeric_bin_definition.get_bin_names(), [
            '...-0]', '[0-1]', '[1-2]', '[2-3]', '[3-4]', '[4-5]', '[5-6]', '[6-7]', '[7-8]', '[8-9]', '[9-10]', '[10-...'])

        linear_numeric_bin_definition = LinearNumericBinDefinition(0, 2, 10, True)
        self.assertEqual(linear_numeric_bin_definition.get_bin_names(), [
            '...-0]', '[0-2]', '[2-4]', '[4-6]', '[6-8]', '[8-10]', '[10-12]', '[12-14]', '[14-16]', '[16-18]', '[18-20]', '[20-...'])

    def test2linear_numeric_bin_definition_get_bin(self):
        """This test case aims to test the functionality of the LinearNumericBinDefinition's get_bin method."""
        linear_numeric_bin_definition = LinearNumericBinDefinition(0, 1, 10, True)
        self.assertEqual(linear_numeric_bin_definition.get_bin(2), 3)

        linear_numeric_bin_definition = LinearNumericBinDefinition(1, 1, 10, True)
        self.assertEqual(linear_numeric_bin_definition.get_bin(2), 2)

        linear_numeric_bin_definition = LinearNumericBinDefinition(2, 1, 10, True)
        self.assertEqual(linear_numeric_bin_definition.get_bin(2), 1)

        linear_numeric_bin_definition = LinearNumericBinDefinition(0, 4, 10, True)
        self.assertEqual(linear_numeric_bin_definition.get_bin(2), 1)

    def test3linear_numeric_bin_definition_get_bin_p_values(self):
        """This test case aims to test the functionality of the LinearNumericBinDefinition's get_bin_p_values method."""
        linear_numeric_bin_definition = LinearNumericBinDefinition(0, 1, 10, True)
        self.assertNotEqual(linear_numeric_bin_definition.get_bin_p_value(2, 10, [2, 2]), None,
                            'Probably the scipy module could not be loaded. Please check your installation.')

    def test4_modulo_time_bin_definition_get_bin(self):
        """This test case aims to test the functionality of the ModuloTimeBinDefinition's getBin method."""
        modulo_time_bin_definition = ModuloTimeBinDefinition(86400, 3600, 0, 1, 24, False)
        self.assertEqual(modulo_time_bin_definition.get_bin(57599), 15)
        self.assertEqual(modulo_time_bin_definition.get_bin(57600), 16)
        self.assertEqual(modulo_time_bin_definition.get_bin(61199), 16)
        self.assertEqual(modulo_time_bin_definition.get_bin(61200), 17)

    def test5_histogram_data_add_value(self):
        """This test case aims to test the addition of Values to HistogramData class."""
        modulo_time_bin_definition = ModuloTimeBinDefinition(86400, 3600, 0, 1, 24, False)
        histogram_data = HistogramData("crontab", modulo_time_bin_definition)
        histogram_data.add_value(57600)
        self.assertEqual(histogram_data.bin_data[16], 1)
        self.assertEqual(histogram_data.total_elements, 1)
        self.assertEqual(histogram_data.binned_elements, 0)

        histogram_data.add_value(61200)
        self.assertEqual(histogram_data.bin_data[16], 1)
        self.assertEqual(histogram_data.bin_data[17], 1)
        self.assertEqual(histogram_data.total_elements, 2)
        self.assertEqual(histogram_data.binned_elements, 0)

        histogram_data.add_value(61500)
        self.assertEqual(histogram_data.bin_data[16], 1)
        self.assertEqual(histogram_data.bin_data[17], 2)
        self.assertEqual(histogram_data.total_elements, 3)
        self.assertEqual(histogram_data.binned_elements, 0)

        histogram_data.add_value(100000)
        # 100000%86400 = 13600 -> 3
        self.assertEqual(histogram_data.bin_data[3], 1)
        self.assertEqual(histogram_data.bin_data[16], 1)
        self.assertEqual(histogram_data.bin_data[17], 2)
        self.assertEqual(histogram_data.total_elements, 4)
        self.assertEqual(histogram_data.binned_elements, 0)

    def test6_histogram_data_reset(self):
        """This test case aims to test resetting the Values of the HistogramData class."""
        modulo_time_bin_definition = ModuloTimeBinDefinition(86400, 3600, 0, 1, 24, False)
        histogram_data = HistogramData("crontab", modulo_time_bin_definition)
        histogram_data.add_value(57600)
        histogram_data.add_value(61200)
        histogram_data.reset()
        self.assertEqual(histogram_data.total_elements, 0)
        self.assertEqual(histogram_data.binned_elements, 0)
        for item in histogram_data.bin_data:
            self.assertEqual(item, 0)

    def test7_histogram_data_clone(self):
        """This test case aims to test cloning a HistogramData object."""
        modulo_time_bin_definition = ModuloTimeBinDefinition(86400, 3600, 0, 1, 24, False)
        histogram_data = HistogramData("crontab", modulo_time_bin_definition)
        histogram_data.add_value(57600)
        histogram_data.add_value(61200)
        clone = histogram_data.clone()

        self.assertEqual(clone.bin_data[16], 1)
        self.assertEqual(clone.bin_data[17], 1)
        self.assertEqual(clone.total_elements, 2)
        self.assertEqual(clone.binned_elements, 0)

        clone.add_value(1)
        self.assertEqual(clone.bin_data[0], 1)
        self.assertEqual(histogram_data.bin_data[0], 0)

    def test8_histogram_data_to_string(self):
        """This test case aims to test the functionality of the HistogramData's to_string method."""
        modulo_time_bin_definition = ModuloTimeBinDefinition(86400, 3600, 0, 1, 24, False)
        histogram_data = HistogramData("crontab", modulo_time_bin_definition)
        histogram_data.add_value(57600)

        histogram_data.add_value(61200)

        clone = histogram_data.clone()

        self.assertEqual(clone.bin_data[16], 1)
        self.assertEqual(clone.bin_data[17], 1)
        self.assertEqual(clone.total_elements, 2)
        self.assertEqual(clone.binned_elements, 0)

        clone.add_value(1)
        self.assertEqual(clone.bin_data[0], 1)
        self.assertEqual(histogram_data.bin_data[0], 0)
        self.assertEqual(clone.to_string(''), 'Property "crontab" (3 elements):\n* [0-1]: 1 (ratio = 3.33e-01, p = 1.20e-01)\n* [16-17]: '
                                              '1 (ratio = 3.33e-01, p = 1.20e-01)\n* [17-18]: 1 (ratio = 3.33e-01, p = 1.20e-01)')

    def test9HistogramAnalysisReceiveAtomNoReport(self):
        """This test case aims to test the functionality of the HistogramAnalysis's receive_atom method, when NO report is expected."""
        description = "Test9HistogramAnalysis"
        start_time = 57600
        end_time = 662600
        diff = 30000

        modulo_time_bin_definition = ModuloTimeBinDefinition(86400, 3600, 0, 1, 24, False)
        histogram_data = HistogramData(self.match_crontab, modulo_time_bin_definition)
        histogram_analysis = HistogramAnalysis(self.aminer_config, [(histogram_data.property_path, modulo_time_bin_definition)], 604800,
                                               [self.stream_printer_event_handler], True, 'Default')
        self.analysis_context.register_component(histogram_analysis, description)
        match_element = MatchElement(self.match_crontab, str(start_time).encode(), start_time, None)

        t = time.time()
        log_atom = LogAtom(histogram_data.bin_data, ParserMatch(match_element), t, histogram_analysis)
        histogram_analysis.receive_atom(log_atom)
        histogram_data.add_value(start_time)
        histogram_data.add_value(end_time)
        match_element = MatchElement(self.match_crontab, str(end_time).encode(), end_time, None)
        log_atom = LogAtom(histogram_data.bin_data, ParserMatch(match_element), t + diff, histogram_analysis)
        histogram_analysis.receive_atom(log_atom)
        self.assertEqual(self.output_stream.getvalue(), '')

        # resetting the outputStream
        start_time = start_time + 3600
        end_time = end_time + 3600
        self.reset_output_stream()
        t = t + diff

        log_atom = LogAtom(histogram_data.bin_data, ParserMatch(match_element), t, histogram_analysis)
        histogram_analysis.receive_atom(log_atom)

        histogram_data.add_value(start_time)
        histogram_data.add_value(end_time)
        match_element = MatchElement(self.match_crontab, str(end_time).encode(), end_time, None)
        log_atom = LogAtom(histogram_data.bin_data, ParserMatch(match_element), t + diff, histogram_analysis)
        histogram_analysis.receive_atom(log_atom)
        self.assertEqual(self.output_stream.getvalue(), '')

    def test10_histogram_analysis_receive_atom_report_expected(self):
        """This test case aims to test the functionality of the HistogramAnalysis's receive_atom method, when A report is expected."""
        description = "Test10HistogramAnalysis"
        start_time = 57600
        end_time = 662600
        diff = 605000

        modulo_time_bin_definition = ModuloTimeBinDefinition(86400, 3600, 0, 1, 24, False)
        histogram_data = HistogramData(self.match_crontab, modulo_time_bin_definition)
        histogram_analysis = HistogramAnalysis(self.aminer_config, [(histogram_data.property_path, modulo_time_bin_definition)], 604800,
                                               [self.stream_printer_event_handler], True, 'Default')
        self.analysis_context.register_component(histogram_analysis, description)
        match_element = MatchElement(self.match_crontab, str(start_time).encode(), start_time, None)

        t = time.time()
        log_atom = LogAtom(histogram_data.bin_data, ParserMatch(match_element), t, histogram_analysis)
        histogram_analysis.receive_atom(log_atom)
        histogram_data.add_value(start_time)
        histogram_data.add_value(end_time)
        match_element = MatchElement(self.match_crontab, str(end_time).encode(), end_time, None)
        log_atom = LogAtom(histogram_data.bin_data, ParserMatch(match_element), t + diff, histogram_analysis)
        histogram_analysis.receive_atom(log_atom)

        self.assertEqual(self.output_stream.getvalue(), self.__expected_string_histogram_analysis % (
            datetime.fromtimestamp(t + diff).strftime(self.datetime_format_string), histogram_analysis.__class__.__name__, description, 2,
            datetime.fromtimestamp(t).strftime(self.datetime_format_string),
            datetime.fromtimestamp(t + diff).strftime(self.datetime_format_string),
            'Property "match/crontab" (2 elements):\n  * [16-17]: 2 (ratio = 1.00e+00, p = 1.74e-03)'))

        # resetting the outputStream
        start_time = start_time + 3600
        end_time = end_time + 3600
        self.reset_output_stream()
        t = t + diff

        log_atom = LogAtom(histogram_data.bin_data, ParserMatch(match_element), t, histogram_analysis)
        histogram_analysis.receive_atom(log_atom)
        histogram_data.add_value(start_time)
        histogram_data.add_value(end_time)
        match_element = MatchElement(self.match_crontab, str(end_time).encode(), end_time, None)

        log_atom = LogAtom(histogram_data.bin_data, ParserMatch(match_element), t + diff, histogram_analysis)
        histogram_analysis.receive_atom(log_atom)

        self.assertEqual(self.output_stream.getvalue(), self.__expected_string_histogram_analysis % (
            datetime.fromtimestamp(t + diff).strftime(self.datetime_format_string), histogram_analysis.__class__.__name__, description, 2,
            datetime.fromtimestamp(t).strftime(self.datetime_format_string),
            datetime.fromtimestamp(t + diff).strftime(self.datetime_format_string),
            'Property "match/crontab" (2 elements):\n  * [16-17]: 1 (ratio = 5.00e-01, p = 8.16e-02)\n  * [17-18]: 1 '
            '(ratio = 5.00e-01, p = 8.16e-02)'))

    def test11_path_dependent_histogram_analysis_no_report(self):
        """
        This test case aims to test the functionality of the PathDependantHistogramAnalysis.receive_atom method.
        No report is expected.
        """
        description = "Test11HistogramAnalysis"
        start_time = 57600
        end_time = 662600
        diff = 30000

        modulo_time_bin_definition = ModuloTimeBinDefinition(86400, 3600, 0, 1, 24, False)
        histogram_data = HistogramData(self.match_crontab, modulo_time_bin_definition)
        path_dependent_histogram_analysis = PathDependentHistogramAnalysis(
            self.aminer_config, histogram_data.property_path, modulo_time_bin_definition, 604800,
            [self.stream_printer_event_handler], True, 'Default')
        self.analysis_context.register_component(path_dependent_histogram_analysis, description)
        match_element = MatchElement(self.match_crontab, str(start_time).encode(), start_time, None)

        t = time.time()
        log_atom = LogAtom(histogram_data.bin_data, ParserMatch(match_element), t, path_dependent_histogram_analysis)
        path_dependent_histogram_analysis.receive_atom(log_atom)
        histogram_data.add_value(start_time)
        histogram_data.add_value(end_time)
        match_element = MatchElement(self.match_crontab, str(end_time).encode(), end_time, None)
        log_atom = LogAtom(histogram_data.bin_data, ParserMatch(match_element), t + diff, path_dependent_histogram_analysis)
        path_dependent_histogram_analysis.receive_atom(log_atom)
        self.assertEqual(self.output_stream.getvalue(), '')

        # resetting the outputStream
        start_time = start_time + 3600
        end_time = end_time + 3600
        self.reset_output_stream()
        t = t + diff

        log_atom = LogAtom(histogram_data.bin_data, ParserMatch(match_element), t, path_dependent_histogram_analysis)
        path_dependent_histogram_analysis.receive_atom(log_atom)
        match_element = MatchElement(self.match_crontab, str(start_time).encode(), start_time, None)
        log_atom = LogAtom(histogram_data.bin_data, ParserMatch(match_element), t, path_dependent_histogram_analysis)
        path_dependent_histogram_analysis.receive_atom(log_atom)
        histogram_data.add_value(start_time)
        histogram_data.add_value(end_time)
        histogram_data.add_value(start_time)
        histogram_data.add_value(end_time)
        match_element = MatchElement(self.match_crontab, str(end_time).encode(), end_time, None)

        log_atom = LogAtom(histogram_data.bin_data, ParserMatch(match_element), t + diff, path_dependent_histogram_analysis)
        path_dependent_histogram_analysis.receive_atom(log_atom)
        match_element = MatchElement(self.match_crontab, str(end_time).encode(), end_time, None)
        log_atom = LogAtom(histogram_data.bin_data, ParserMatch(match_element), t + diff, path_dependent_histogram_analysis)
        path_dependent_histogram_analysis.receive_atom(log_atom)
        self.assertEqual(self.output_stream.getvalue(), '')

    def test12_path_dependent_histogram_analysis_report_expected(self):
        """
        This test case aims to test the functionality of the PathDependantHistogramAnalysis.receiveAtom method.
        A report is expected.
        """
        description = "Test12HistogramAnalysis"
        start_time = 57600
        end_time = 662600
        diff = 605000

        modulo_time_bin_definition = ModuloTimeBinDefinition(86400, 3600, 0, 1, 24, False)
        histogram_data = HistogramData(self.match_crontab, modulo_time_bin_definition)
        path_dependent_histogram_analysis = PathDependentHistogramAnalysis(
            self.aminer_config, histogram_data.property_path, modulo_time_bin_definition, 604800,
            [self.stream_printer_event_handler], True, 'Default')
        self.analysis_context.register_component(path_dependent_histogram_analysis, description)
        match_element = MatchElement(self.match_crontab, str(start_time).encode(), start_time, None)

        t = time.time()
        log_atom = LogAtom(histogram_data.bin_data, ParserMatch(match_element), t, path_dependent_histogram_analysis)
        path_dependent_histogram_analysis.receive_atom(log_atom)
        histogram_data.add_value(start_time)
        histogram_data.add_value(end_time)
        match_element = MatchElement(self.match_crontab, str(end_time).encode(), end_time, None)
        log_atom = LogAtom(histogram_data.bin_data, ParserMatch(match_element), t + diff, path_dependent_histogram_analysis)
        path_dependent_histogram_analysis.receive_atom(log_atom)
        self.assertEqual(self.output_stream.getvalue(), self.__expected_string_path_dependent_histogram_analysis % (
            datetime.fromtimestamp(t + diff).strftime(self.datetime_format_string), path_dependent_histogram_analysis.__class__.__name__,
            description, 2, datetime.fromtimestamp(t).strftime(self.datetime_format_string),
            datetime.fromtimestamp(t + diff).strftime(self.datetime_format_string),
            'Path values "match/crontab":\nExample: 662600\n  Property "match/crontab" (2 elements):\n  * [16-17]: 2 '
            '(ratio = 1.00e+00, p = 1.74e-03)'))

        # resetting the outputStream
        start_time = start_time + 3600
        end_time = end_time + 3600
        self.output_stream.seek(0)
        self.output_stream.truncate(0)
        t = t + diff

        log_atom = LogAtom(histogram_data.bin_data, ParserMatch(match_element), t, path_dependent_histogram_analysis)
        path_dependent_histogram_analysis.receive_atom(log_atom)
        match_element = MatchElement(self.match_crontab, str(start_time).encode(), start_time, None)
        log_atom = LogAtom(histogram_data.bin_data, ParserMatch(match_element), t, path_dependent_histogram_analysis)
        path_dependent_histogram_analysis.receive_atom(log_atom)
        histogram_data.add_value(start_time)
        histogram_data.add_value(end_time)
        histogram_data.add_value(start_time)
        histogram_data.add_value(end_time)
        match_element = MatchElement(self.match_crontab, str(end_time).encode(), end_time, None)

        log_atom = LogAtom(histogram_data.bin_data, ParserMatch(match_element), t + diff, path_dependent_histogram_analysis)
        path_dependent_histogram_analysis.receive_atom(log_atom)
        match_element = MatchElement(self.match_crontab, str(end_time).encode(), end_time, None)
        log_atom = LogAtom(histogram_data.bin_data, ParserMatch(match_element), t + diff, path_dependent_histogram_analysis)
        path_dependent_histogram_analysis.receive_atom(log_atom)
        self.assertEqual(self.output_stream.getvalue(), self.__expected_string_path_dependent_histogram_analysis % (
            datetime.fromtimestamp(t + diff).strftime(self.datetime_format_string), path_dependent_histogram_analysis.__class__.__name__,
            description, 3, datetime.fromtimestamp(t).strftime(self.datetime_format_string),
            datetime.fromtimestamp(t + diff).strftime(self.datetime_format_string),
            'Path values "match/crontab":\nExample: 666200\n  Property "match/crontab" (3 elements):\n  * [16-17]: 1 '
            '(ratio = 3.33e-01, p = 1.20e-01)\n  * [17-18]: 2 (ratio = 6.67e-01, p = 5.06e-03)'))


if __name__ == "__main__":
    unittest.main()



"""
This module defines a writer that forwards match information to a stream.

This program is free software: you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.
This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
You should have received a copy of the GNU General Public License along with
this program. If not, see <http://www.gnu.org/licenses/>.
"""

from aminer.AnalysisChild import AnalysisContext
from aminer.input.InputInterfaces import AtomHandlerInterface
from aminer.util.TimeTriggeredComponentInterface import TimeTriggeredComponentInterface
import _io


class MatchValueStreamWriter(AtomHandlerInterface, TimeTriggeredComponentInterface):
    """
    This class extracts values from a given match and writes them to a stream.
    This can be used to forward these values to another program (when stream is a wrapped network socket) or to a file for further analysis.
    A stream is used instead of a file descriptor to increase performance. To flush it from time to time, add the writer object also to the
    time trigger list.
    """

    time_trigger_class = AnalysisContext.TIME_TRIGGER_CLASS_REALTIME

    def __init__(self, stream, target_path_list, separator, missing_value_string):
        """
        Initialize the writer.
        @param stream the stream on which the match results are written.
        @param target_path_list parser paths of values to be analyzed. Multiple paths mean that all values occurring in these paths are
               considered for value range generation.
        @param separator a string to be added between match values in the output stream.
        @param missing_value_string a string which is added if no match was found.
        """
        # avoid "defined outside init" issue
        self.log_success, self.log_total = [None]*2
        super().__init__(stream=stream, target_path_list=target_path_list, separator=separator, missing_value_string=missing_value_string)

    def receive_atom(self, log_atom):
        """Forward match value information to the stream."""
        self.log_total += 1
        match_dict = log_atom.parser_match.get_match_dictionary()
        add_sep_flag = False
        contains_data = False
        result = b''
        for path in self.target_path_list:
            if add_sep_flag:
                result += self.separator
            match = match_dict.get(path)
            if match is None:
                result += self.missing_value_string
            else:
                matches = []
                if isinstance(match, list):
                    matches = match
                else:
                    matches.append(match)
                for match in matches:
                    result += match.match_string + self.separator
                    contains_data = True
                if len(self.separator) > 0:
                    result = result[:-len(self.separator)]
            add_sep_flag = True
        if contains_data:
            if not isinstance(self.stream, _io.BytesIO):
                self.stream.write(result.decode('ascii', 'ignore'))
                self.stream.write('\n')
            else:
                self.stream.write(result)
                self.stream.write(b'\n')
            self.log_success += 1

    def do_timer(self, _trigger_time):
        """Flush the timer."""
        self.stream.flush()
        return 10

    def do_persist(self):
        """Flush the timer."""
        self.stream.flush()



"""
This module defines a model element to detect fixed strings from a list of words.

This program is free software: you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.
This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
You should have received a copy of the GNU General Public License along with
this program. If not, see <http://www.gnu.org/licenses/>.
"""

from aminer.parsing.ModelElementInterface import ModelElementInterface
from aminer.parsing.MatchElement import MatchElement


class FixedWordlistDataModelElement(ModelElementInterface):
    """
    This class defines a model element to detect fixed strings from a list of words.
    The match will return the position of the word in the search list, thus the sorting of the list is important. Apart from that, the
    wordlist must not contain any words, that are identical to the beginning of words later in the list. In that case, the longer match
    could never be detected.
    """

    def __init__(self, element_id: str, wordlist: list):
        """
        Create the model element.
        @param wordlist the list of words to search for. If it does not fulfill the sorting criteria mentioned in the class documentation,
               an Exception will be raised.
        """
        super().__init__(element_id, wordlist=wordlist)

    def get_match_element(self, path: str, match_context):
        """@return None when there is no match, MatchElement otherwise."""
        data = match_context.match_data
        match_data = None
        word_pos = 0
        for word in self.wordlist:
            if data.startswith(word):
                match_data = word
                break
            word_pos += 1

        if match_data is None:
            return None

        match_context.update(match_data)
        return MatchElement(f"{path}/{self.element_id}", match_data, word_pos, None)



import unittest
from aminer.analysis.EntropyDetector import EntropyDetector
from aminer.input.LogAtom import LogAtom
from aminer.parsing.MatchElement import MatchElement
from aminer.parsing.ParserMatch import ParserMatch
from unit.TestBase import TestBase


class TestHandler():
    """Dummy anomaly handler."""

    def __init__(self):
        self.anomaly = None

    # skipcq: PYL-W0613
    def receive_event(self, name, msg, ll, evdat, atom, obj):
        """Receive anomaly information."""
        self.anomaly = evdat


class EntropyDetectorTest(TestBase):
    """Unittests for the EntropyDetector."""

    def test1_normal_sequence_detection(self):
        """
        This test case checks the normal detection of new character sets.
        The charset detector is used to learn an alphabet and detect new characters for different identifiers.
        """
        description = "Test1EntropyDetector"

        # Initialize detector
        test_handler = TestHandler()
        event_entropy_detector = EntropyDetector(self.aminer_config, [test_handler], ['/value'], 0.05, False, False, 'Default', True,
                                                 False)
        self.analysis_context.register_component(event_entropy_detector, description)

        # Prepare log atoms that represent string values. Anomalies are detected when character pair distributions deviate.
        # The following events are generated:
        #  value: aminer
        #  value: logdata-anomaly-miner
        #  value: ait-aecid
        #  value: austrian
        #  value: institute
        #  value: lfmvasacz
        m_1 = MatchElement('/value', b'aminer', b'aminer', None)
        parser_match_1 = ParserMatch(m_1)
        log_atom_1 = LogAtom(b'aminer', parser_match_1, 1, None)

        m_2 = MatchElement('/value', b'logdata-anomaly-miner', b'logdata-anomaly-miner', None)
        parser_match_2 = ParserMatch(m_2)
        log_atom_2 = LogAtom(b'logdata-anomaly-miner', parser_match_2, 2, None)

        m_3 = MatchElement('/value', b'ait-aecid', b'ait-aecid', None)
        parser_match_3 = ParserMatch(m_3)
        log_atom_3 = LogAtom(b'ait-aecid', parser_match_3, 3, None)

        m_4 = MatchElement('/value', b'austrian', b'austrian', None)
        parser_match_4 = ParserMatch(m_4)
        log_atom_4 = LogAtom(b'austrian', parser_match_4, 4, None)

        m_5 = MatchElement('/value', b'institute', b'institute', None)
        parser_match_5 = ParserMatch(m_5)
        log_atom_5 = LogAtom(b'institute', parser_match_5, 5, None)

        m_6 = MatchElement('/value', b'lfmvasacz', b'lfmvasacz', None)
        parser_match_6 = ParserMatch(m_6)
        log_atom_6 = LogAtom(b'lfmvasacz', parser_match_6, 6, None)

        # Forward log atoms to detector
        # First value should generate an anomaly, because no frequencies are known yet
        # Input: aminer
        # Expected output: Anomaly
        event_entropy_detector.receive_atom(log_atom_1)
        self.assertEqual(test_handler.anomaly, {'AnalysisComponent': {'AffectedLogAtomPaths': ['/value'],
                                                                      'AffectedLogAtomValues': ['aminer'],
                                                                      'CriticalValue': 0.0,
                                                                      'ProbabilityThreshold': 0.05}})
        test_handler.anomaly = None

        # Second value should not generate an anomaly, because it contains substring 'miner' which shares charpairs with 'aminer'
        # Input: logdata-anomaly-miner
        # Expected output: None
        event_entropy_detector.receive_atom(log_atom_2)
        self.assertIsNone(test_handler.anomaly)

        # Third value should not generate an anomaly, since it is a normal string
        # Input: ait-aecid
        # Expected output: None
        event_entropy_detector.receive_atom(log_atom_3)
        self.assertIsNone(test_handler.anomaly)

        # Fourth value should not generate an anomaly, since it is a normal string
        # Input: austrian
        # Expected output: None
        event_entropy_detector.receive_atom(log_atom_4)
        self.assertIsNone(test_handler.anomaly)

        # Fifth value should not generate an anomaly, since it is a normal string
        # Input: institute
        # Expected output: None
        event_entropy_detector.receive_atom(log_atom_5)
        self.assertIsNone(test_handler.anomaly)

        # Sixth value should generate an anomaly, since it is a randomly generated string
        # Input: lfmvasacz
        # Expected output: Anomaly
        event_entropy_detector.receive_atom(log_atom_6)
        self.assertEqual(test_handler.anomaly, {'AnalysisComponent': {'AffectedLogAtomPaths': ['/value'],
                                                                      'AffectedLogAtomValues': ['lfmvasacz'],
                                                                      'CriticalValue': 0.02,
                                                                      'ProbabilityThreshold': 0.05}})
        test_handler.anomaly = None


if __name__ == "__main__":
    unittest.main()



"""This module defines the parsing model for ntpd logs."""

from aminer.parsing.DateTimeModelElement import DateTimeModelElement
from aminer.parsing.DecimalFloatValueModelElement import DecimalFloatValueModelElement
from aminer.parsing.DecimalIntegerValueModelElement import DecimalIntegerValueModelElement
from aminer.parsing.DelimitedDataModelElement import DelimitedDataModelElement
from aminer.parsing.FirstMatchModelElement import FirstMatchModelElement
from aminer.parsing.FixedDataModelElement import FixedDataModelElement
from aminer.parsing.IpAddressDataModelElement import IpAddressDataModelElement
from aminer.parsing.SequenceModelElement import SequenceModelElement
from aminer.parsing.VariableByteDataModelElement import VariableByteDataModelElement


def get_model():
    """Get the model."""
    interface_name_model = VariableByteDataModelElement("interface", b"0123456789abcdefghijklmnopqrstuvwxyz.")
    dtme = DateTimeModelElement("expire-date", b"%Y-%m-%dT%H:%M:%SZ")

    type_children = [
        SequenceModelElement("exit", [
            FixedDataModelElement("s0", b"ntpd exiting on signal "),
            DecimalIntegerValueModelElement("signal")
        ]),
        SequenceModelElement("listen-drop", [
            FixedDataModelElement("s0", b"Listen and drop on "),
            DecimalIntegerValueModelElement("fd"),
            FixedDataModelElement("s1", b" "),
            interface_name_model,
            FixedDataModelElement("s2", b" "),
            FirstMatchModelElement("address", [
                IpAddressDataModelElement("ipv4"),
                DelimitedDataModelElement("ipv6", b" "),
                FixedDataModelElement("ipv6_missing", b"[::]")
            ]),
            FirstMatchModelElement("udp", [
                FixedDataModelElement("s3", b" UDP 123"),
                FixedDataModelElement("s3", b":123")])
        ]),
        SequenceModelElement("listen-normal", [
            FixedDataModelElement("s0", b"Listen normally on "),
            DecimalIntegerValueModelElement("fd"),
            FixedDataModelElement("s1", b" "),
            interface_name_model,
            FixedDataModelElement("s2", b" "),
            IpAddressDataModelElement("ip"),
            FirstMatchModelElement("msg", [
                FixedDataModelElement("port-new", b":123"),
                FixedDataModelElement("port-old", b" UDP 123")
            ])
        ]),
        SequenceModelElement("listen-routing", [
            FixedDataModelElement("s0", b"Listening on routing socket on fd #"),
            DecimalIntegerValueModelElement("fd"),
            FixedDataModelElement("s1", b" for interface updates")
        ]),
        SequenceModelElement("soliciting-pool", [
            FixedDataModelElement("s0", b"Soliciting pool server "),
            IpAddressDataModelElement("pool-server-ip")
        ]),
        SequenceModelElement("starting", [
            FixedDataModelElement("s0", b"ntpd "),
            DelimitedDataModelElement("version", b" "),
            FixedDataModelElement("s1", b" (1): Starting")
        ]),
        SequenceModelElement("no-root", [
            FixedDataModelElement("s0", b"must be run as root, not uid "),
            DecimalIntegerValueModelElement("uid")
        ]),
        SequenceModelElement("leapsecond-file", [
            FixedDataModelElement("s0", b"leapsecond file ('"),
            DelimitedDataModelElement("file", b"'"),
            FixedDataModelElement("s1", b"'): "),
            FirstMatchModelElement("first", [
                FixedDataModelElement("msg", b"good hash signature"),
                SequenceModelElement("seq", [
                    FixedDataModelElement("s2", b"loaded, expire="),
                    dtme,
                    FixedDataModelElement("s3", b" last="),
                    dtme,
                    FixedDataModelElement("s4", b" ofs="),
                    DecimalIntegerValueModelElement("ofs")
                ])
            ])
        ]),
        FixedDataModelElement("unable-to-bind", b"unable to bind to wildcard address :: - another process may be running - EXITING"),
        FixedDataModelElement("new-interfaces", b"new interface(s) found: waking up resolver"),
        FixedDataModelElement("ntp-io", b"ntp_io: estimated max descriptors: 1024, initial socket boundary: 16"),
        FixedDataModelElement("peers-refreshed", b"peers refreshed"),
        FixedDataModelElement("log-file", b"logging to file /var/log/ntplog"),
        FixedDataModelElement("command-line", b"Command line: ntpd"),
        SequenceModelElement("precision", [
            FixedDataModelElement("s0", b"proto: precision = "),
            DecimalFloatValueModelElement("precision"),
            FixedDataModelElement("s1", b" usec ("),
            DecimalIntegerValueModelElement("usec", value_sign_type=DecimalIntegerValueModelElement.SIGN_TYPE_OPTIONAL),
            FixedDataModelElement("s2", b")")
        ])]

    model = SequenceModelElement("ntpd", [
        FixedDataModelElement("sname", b"ntpd["),
        DecimalIntegerValueModelElement("pid"),
        FixedDataModelElement("s0", b"]: "),
        FirstMatchModelElement("msg", type_children)
    ])
    return model



# DISCLAIMER: adapted code from
# https://stackoverflow.com/questions/6886283/how-i-can-i-lazily-read-multiple-json-values-from-a-file-stream-in-python

# A streaming byte oriented JSON parser.  Feed it a single byte at a time and
# it will emit complete objects as it comes across them.  Whitespace within and
# between objects is ignored.  This means it can parse newline delimited JSON.
import math


def json_machine(emit, next_func=None):  # skipcq: PY-D0003
    def _value(byte_data):  # skipcq: PY-D0003
        if not byte_data:
            return None

        if byte_data in (0x09, 0x0a, 0x0d, 0x20):
            return _value  # Ignore whitespace

        # only allow json objects in our case
        if byte_data != 0x7b and next_func is _value:
            return None

        if byte_data == 0x22:  # "
            return string_machine(on_value)

        if byte_data in (0x2b, 0x2d) or (0x30 <= byte_data < 0x3a):  # -, + or 0-9
            return number_machine(byte_data, on_number)

        if byte_data == 0x7b:  #:
            return object_machine(on_value)

        if byte_data == 0x5b:  # [
            return array_machine(on_value)

        if byte_data == 0x74:  # t
            return constant_machine(TRUE, True, on_value)

        if byte_data == 0x66:  # f
            return constant_machine(FALSE, False, on_value)

        if byte_data == 0x6e:  # n
            return constant_machine(NULL, None, on_value)

        if next_func is _value:
            return None

        return next_func(byte_data)

    def on_value(value):  # skipcq: PY-D0003, PTC-W0065
        emit(value)
        return next_func

    def on_number(number, byte):  # skipcq: PY-D0003, PTC-W0065
        emit(number)
        return _value(byte)

    next_func = next_func or _value
    return _value


TRUE = [0x72, 0x75, 0x65]
FALSE = [0x61, 0x6c, 0x73, 0x65]
NULL = [0x75, 0x6c, 0x6c]


def constant_machine(bytes_data, value, emit):  # skipcq: PY-D0003
    i = 0
    length = len(bytes_data)

    def _constant(byte_data):  # skipcq: PY-D0003
        nonlocal i
        if byte_data != bytes_data[i]:
            i += 1
            return None

        i += 1
        if i < length:
            return _constant
        return emit(value)

    return _constant


def string_machine(emit):  # skipcq: PY-D0003
    string = ""

    def _string(byte_data):  # skipcq: PY-D0003
        nonlocal string

        if byte_data == 0x22:  # "
            return emit(string)

        if byte_data == 0x5c:  # \
            return _escaped_string

        if byte_data & 0x80:  # UTF-8 handling
            return utf8_machine(byte_data, on_char_code)

        if byte_data < 0x20 and byte_data != 0xa:  # ASCII control character - \n is allowed
            return None

        string += chr(byte_data)
        return _string

    def _escaped_string(byte_data):  # skipcq: PY-D0003, PTC-W0065
        nonlocal string

        if byte_data in (0x22, 0x5c, 0x2f):  # " \ /
            string += chr(byte_data)
            return _string

        if byte_data == 0x62:  # b
            string += "\b"
            return _string

        if byte_data == 0x66:  # f
            string += "\f"
            return _string

        if byte_data == 0x6e:  # n
            string += "\n"
            return _string

        if byte_data == 0x72:  # r
            string += "\r"
            return _string

        if byte_data == 0x74:  # t
            string += "\t"
            return _string

        if byte_data == 0x75:  # u
            return hex_machine(on_char_code)

        return None

    def on_char_code(char_code):  # skipcq: PY-D0003, PTC-W0065
        nonlocal string
        string += chr(char_code)
        return _string

    return _string


# Nestable state machine for UTF-8 Decoding.
def utf8_machine(byte_data, emit):  # skipcq: PY-D0003
    left = 0
    num = 0

    def _utf8(byte_data):  # skipcq: PY-D0003
        nonlocal num, left
        if (byte_data & 0xc0) != 0x80:
            return None

        left = left - 1

        num |= (byte_data & 0x3f) << (left * 6)
        if left:
            return _utf8
        return emit(num)

    if 0xc0 <= byte_data < 0xe0:  # 2-byte UTF-8 Character
        left = 1
        num = (byte_data & 0x1f) << 6
        return _utf8

    if 0xe0 <= byte_data < 0xf0:  # 3-byte UTF-8 Character
        left = 2
        num = (byte_data & 0xf) << 12
        return _utf8

    if 0xf0 <= byte_data < 0xf8:  # 4-byte UTF-8 Character
        left = 3
        num = (byte_data & 0x07) << 18
        return _utf8

    return None


# Nestable state machine for hex escaped characters
def hex_machine(emit):  # skipcq: PY-D0003
    left = 4
    num = 0

    def _hex(byte_data):  # skipcq: PY-D0003
        nonlocal num, left

        if 0x30 <= byte_data < 0x3a:
            i = byte_data - 0x30
        elif 0x61 <= byte_data <= 0x66:
            i = byte_data - 0x57
        elif 0x41 <= byte_data <= 0x46:
            i = byte_data - 0x37
        else:
            return None

        left -= 1
        num |= i << (left * 4)

        if left:
            return _hex
        return emit(num)

    return _hex


def number_machine(byte_data, emit):  # skipcq: PY-D0003
    sign = 1
    number = 0
    decimal = 0
    esign = 1
    exponent = 0
    dividend = 10
    start_with_zero = False

    def _mid(byte_data):  # skipcq: PY-D0003
        if start_with_zero and byte_data not in (0x2e, 0x45, 0x65, 0x7d, 0x2c, 0xa, 0x20):  # . E e } , \n Space
            return None

        if byte_data == 0x2e:  # .
            return _decimal

        return _later(byte_data)

    def _number(byte_data):  # skipcq: PY-D0003
        nonlocal number
        if 0x30 <= byte_data < 0x3a:
            number = number * 10 + (byte_data - 0x30)
            return _number

        return _mid(byte_data)

    def _start(byte_data):  # skipcq: PY-D0003
        nonlocal start_with_zero
        if byte_data == 0x30:
            start_with_zero = True
            return _mid

        if 0x30 < byte_data < 0x3a:
            return _number(byte_data)

        return None

    def _decimal(byte_data):  # skipcq: PY-D0003
        nonlocal decimal
        nonlocal dividend
        if 0x30 <= byte_data < 0x3a:
            decimal += (byte_data - 0x30) / dividend
            dividend *= 10
            return _decimal

        return _later(byte_data)

    def _later(byte_data):  # skipcq: PY-D0003, PTC-W0065
        if byte_data in (0x45, 0x65):  # E e
            return _esign

        return _done(byte_data)

    def _esign(byte_data):  # skipcq: PY-D0003, PTC-W0065
        nonlocal esign
        if byte_data == 0x2b:  # +
            return _exponent

        if byte_data == 0x2d:  # -
            esign = -1
            return _exponent

        return _exponent(byte_data)

    def _exponent(byte_data):  # skipcq: PY-D0003
        nonlocal exponent
        if 0x30 <= byte_data < 0x3a:
            exponent = exponent * 10 + (byte_data - 0x30)
            return _exponent

        return _done(byte_data)

    def _done(byte_data):  # skipcq: PY-D0003, PTC-W0065
        value = sign * (number + decimal)
        if exponent:
            value *= math.pow(10, esign * exponent)

        return emit(value, byte_data)

    if byte_data == 0x2d:  # -
        sign = -1
        return _start

    if byte_data == 0x2b:  # +
        return _start

    return _start(byte_data)


def array_machine(emit):  # skipcq: PY-D0003
    array_data = []

    def _array(byte_data):  # skipcq: PY-D0003
        if byte_data == 0x5d:  # ]
            return emit(array_data)

        return json_machine(on_value, _comma)(byte_data)

    def on_value(value):  # skipcq: PY-D0003
        array_data.append(value)

    def _comma(byte_data):  # skipcq: PY-D0003
        if byte_data in (0x09, 0x0a, 0x0d, 0x20):
            return _comma  # Ignore whitespace

        if byte_data == 0x2c:  # ,
            return json_machine(on_value, _comma)

        if byte_data == 0x5d:  # ]
            return emit(array_data)

        return None

    return _array


def object_machine(emit):  # skipcq: PY-D0003
    object_data = {}
    key = None

    def _object(byte_data):  # skipcq: PY-D0003
        if byte_data == 0x7d:  #
            return emit(object_data)

        return _key(byte_data)

    def _key(byte_data):  # skipcq: PY-D0003
        if byte_data in (0x09, 0x0a, 0x0d, 0x20):
            return _object  # Ignore whitespace

        if byte_data == 0x22:
            return string_machine(on_key)

        return None

    def on_key(result):  # skipcq: PY-D0003, PTC-W0065
        nonlocal key
        key = result
        return _colon

    def _colon(byte_data):  # skipcq: PY-D0003
        if byte_data in (0x09, 0x0a, 0x0d, 0x20):
            return _colon  # Ignore whitespace

        if byte_data == 0x3a:  # :
            return json_machine(on_value, _comma)

        return None

    def on_value(value):  # skipcq: PY-D0003, PTC-W0065
        object_data[key] = value

    def _comma(byte_data):  # skipcq: PY-D0003
        if byte_data in (0x09, 0x0a, 0x0d, 0x20):
            return _comma  # Ignore whitespace

        if byte_data == 0x2c:  # ,
            return _key

        if byte_data == 0x7d:  #
            return emit(object_data)

        return None

    return _object



"""
This is the interface-class for the TimeTriggeredComponent.

This program is free software: you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.
This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
You should have received a copy of the GNU General Public License along with
this program. If not, see <http://www.gnu.org/licenses/>.

"""

import abc


class TimeTriggeredComponentInterface(metaclass=abc.ABCMeta):
    """
    This is the common interface of all components that can be registered to receive timer interrupts.
    There might be different timelines for triggering, real time and normalized log data time scale for forensic analysis. For forensic
    analyis different timers might be available to register a component. Therefore the component should state, which type of triggering it
    would require.
    """

    @property
    @abc.abstractmethod
    def time_trigger_class(self):
        raise NotImplementedError

    def get_time_trigger_class(self):
        """
        Get the trigger class this component can be registered for.
        See AnalysisContext class for different trigger classes available.
        """
        if self.time_trigger_class not in (1, 2):
            raise NotImplementedError("The self.time_trigger_class property must be set to AnalysisContext.TIME_TRIGGER_CLASS_REALTIME or "
                                      "AnalysisContext.TIME_TRIGGER_CLASS_ANALYSISTIME.")
        return self.time_trigger_class

    @abc.abstractmethod
    def do_timer(self, trigger_time):
        """
        Perform trigger actions and to determine the time for next invocation.
        The caller may decide to invoke this method earlier than requested during the previous call. Classes implementing this method have
        to handle such cases. Each class should try to limit the time spent in this method as it might delay trigger signals to other
        components. For extensive compuational work or IO, a separate thread should be used.
        @param trigger_time the time this trigger is invoked. This might be the current real time when invoked from real time
        timers or the forensic log timescale time value.
        @return the number of seconds when next invocation of this trigger is required.
        """



"""
This file defines the basic NewMatchPathValueComboDetector detector.
It extracts values from LogAtoms and check, if the value combination was already seen before.

This program is free software: you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.
This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
You should have received a copy of the GNU General Public License along with
this program. If not, see <http://www.gnu.org/licenses/>.
"""

import time
import os
import logging

from aminer.AminerConfig import build_persistence_file_name, DEBUG_LOG_NAME, KEY_PERSISTENCE_PERIOD, DEFAULT_PERSISTENCE_PERIOD,\
    STAT_LOG_NAME, CONFIG_KEY_LOG_LINE_PREFIX, DEFAULT_LOG_LINE_PREFIX
from aminer import AminerConfig
from aminer.AnalysisChild import AnalysisContext
from aminer.events.EventInterfaces import EventSourceInterface
from aminer.input.InputInterfaces import AtomHandlerInterface
from aminer.util import PersistenceUtil
from aminer.util.TimeTriggeredComponentInterface import TimeTriggeredComponentInterface


class NewMatchPathValueComboDetector(AtomHandlerInterface, TimeTriggeredComponentInterface, EventSourceInterface):
    """This class creates events when a new value combination for a given list of match data pathes were found."""

    time_trigger_class = AnalysisContext.TIME_TRIGGER_CLASS_REALTIME

    def __init__(self, aminer_config, target_path_list, anomaly_event_handlers, persistence_id='Default', allow_missing_values_flag=False,
                 learn_mode=False, output_logline=True, stop_learning_time=None, stop_learning_no_anomaly_time=None):
        """
        Initialize the detector. This will also trigger reading or creation of persistence storage location.
        @param aminer_config configuration from analysis_context.
        @param target_path_list parser paths of values to be analyzed. Multiple paths mean that all values occurring in these paths are
               considered for value range generation.
        @param anomaly_event_handlers for handling events, e.g., print events to stdout.
        @param persistence_id name of persistence file.
        @param allow_missing_values_flag when set to True, the detector will also use matches, where one of the paths from target_path_list
        does not refer to an existing parsed data object.
        @param learn_mode when set to True, this detector will report a new value only the first time before including it
        in the known values set automatically.
        @param output_logline specifies whether the full parsed log atom should be provided in the output.
        @param stop_learning_time switch the learn_mode to False after the time.
        @param stop_learning_no_anomaly_time switch the learn_mode to False after no anomaly was detected for that time.
        """
        # avoid "defined outside init" issue
        self.learn_mode, self.stop_learning_timestamp, self.next_persist_time, self.log_success, self.log_total = [None]*5
        super().__init__(
            aminer_config=aminer_config, target_path_list=target_path_list, anomaly_event_handlers=anomaly_event_handlers,
            persistence_id=persistence_id, allow_missing_values_flag=allow_missing_values_flag, learn_mode=learn_mode,
            output_logline=output_logline, stop_learning_time=stop_learning_time,
            stop_learning_no_anomaly_time=stop_learning_no_anomaly_time
        )
        self.log_learned_path_value_combos = 0
        self.log_new_learned_values = []

        self.persistence_file_name = build_persistence_file_name(aminer_config, self.__class__.__name__, persistence_id)
        self.known_values_set = set()
        self.load_persistence_data()
        PersistenceUtil.add_persistable_component(self)

    def load_persistence_data(self):
        """Load the persistence data from storage."""
        persistence_data = PersistenceUtil.load_json(self.persistence_file_name)
        if persistence_data is not None:
            # Set and tuples were stored as list of lists. Transform the inner lists to tuples to allow hash operation needed by set.
            self.known_values_set = {tuple(record) for record in persistence_data}
            logging.getLogger(DEBUG_LOG_NAME).debug(f'{self.__class__.__name__} loaded persistence data.')

    def receive_atom(self, log_atom):
        """
        Receive on parsed atom and the information about the parser match.
        @return True if a value combination was extracted and checked against the list of known combinations, no matter if the checked
        values were new or not.
        """
        self.log_total += 1
        if self.learn_mode is True and self.stop_learning_timestamp is not None and \
                self.stop_learning_timestamp < log_atom.atom_time:
            logging.getLogger(DEBUG_LOG_NAME).info(f"Stopping learning in the {self.__class__.__name__}.")
            self.learn_mode = False
        match_dict = log_atom.parser_match.get_match_dictionary()
        match_value_list = []
        for target_path in self.target_path_list:
            match_element = match_dict.get(target_path)
            if match_element is None:
                if not self.allow_missing_values_flag:
                    return False
                match_value_list.append(None)
            else:
                matches = []
                if isinstance(match_element, list):
                    matches = match_element
                else:
                    matches.append(match_element)
                for match_element in matches:
                    match_value_list.append(match_element.match_object)

        match_value_tuple = tuple(match_value_list)
        affected_log_atom_values = []
        for match_value in match_value_list:
            if isinstance(match_value, bytes):
                match_value = match_value.decode(AminerConfig.ENCODING)
            affected_log_atom_values.append(str(match_value))
        if match_value_tuple not in self.known_values_set:
            if self.learn_mode:
                self.known_values_set.add(match_value_tuple)
                self.log_learned_path_value_combos += 1
                self.log_new_learned_values.append(match_value_tuple)
                if self.stop_learning_timestamp is not None and self.stop_learning_no_anomaly_time is not None:
                    self.stop_learning_timestamp = time.time() + self.stop_learning_no_anomaly_time

            analysis_component = {'AffectedLogAtomPaths': self.target_path_list, 'AffectedLogAtomValues': affected_log_atom_values}
            event_data = {'AnalysisComponent': analysis_component}
            try:
                data = log_atom.raw_data.decode(AminerConfig.ENCODING)
            except UnicodeError:
                data = repr(log_atom.raw_data)
            original_log_line_prefix = self.aminer_config.config_properties.get(CONFIG_KEY_LOG_LINE_PREFIX, DEFAULT_LOG_LINE_PREFIX)
            sorted_log_lines = [str(match_value_tuple) + os.linesep + original_log_line_prefix + data]
            for listener in self.anomaly_event_handlers:
                listener.receive_event(f'Analysis.{self.__class__.__name__}', 'New value combination(s) detected', sorted_log_lines,
                                       event_data, log_atom, self)
        self.log_success += 1
        return True

    def do_timer(self, trigger_time):
        """Check if current ruleset should be persisted."""
        if self.next_persist_time is None:
            return self.aminer_config.config_properties.get(KEY_PERSISTENCE_PERIOD, DEFAULT_PERSISTENCE_PERIOD)

        delta = self.next_persist_time - trigger_time
        if delta <= 0:
            self.do_persist()
            delta = self.aminer_config.config_properties.get(KEY_PERSISTENCE_PERIOD, DEFAULT_PERSISTENCE_PERIOD)
            self.next_persist_time = time.time() + delta
        return delta

    def do_persist(self):
        """Immediately write persistence data to storage."""
        PersistenceUtil.store_json(self.persistence_file_name, list(self.known_values_set))
        logging.getLogger(DEBUG_LOG_NAME).debug(f'{self.__class__.__name__} persisted data.')

    def allowlist_event(self, event_type, event_data, allowlisting_data):
        """
        Allowlist an event generated by this source using the information emitted when generating the event.
        @return a message with information about allowlisting
        @throws Exception when allowlisting of this special event using given allowlisting_data was not possible.
        """
        if event_type != f'Analysis.{self.__class__.__name__}':
            msg = 'Event not from this source'
            logging.getLogger(DEBUG_LOG_NAME).error(msg)
            raise Exception(msg)
        if allowlisting_data is not None:
            msg = 'Allowlisting data not understood by this detector'
            logging.getLogger(DEBUG_LOG_NAME).error(msg)
            raise Exception(msg)
        self.known_values_set.add(event_data)
        return f"Allowlisted path(es) {', '.join(self.target_path_list)} with {event_data}."

    def add_to_persistency_event(self, event_type, event_data):
        """
        Add or overwrite the information of event_data to the persistence of component_name.
        @return a message with information about the addition to the persistence.
        @throws Exception when the addition of this special event using given event_data was not possible.
        """
        if event_type != f'Analysis.{self.__class__.__name__}':
            msg = 'Event not from this source'
            logging.getLogger(DEBUG_LOG_NAME).error(msg)
            raise Exception(msg)
        if not isinstance(event_data, list) or len(event_data) != len(self.target_path_list):
            msg = 'Event_data has the wrong format.' \
                'The supported format is [value_1, value_2, ..., value_n] where n is the number of analyzed paths.'
            logging.getLogger(DEBUG_LOG_NAME).error(msg)
            raise Exception(msg)

        match_value_list = []
        for match_element in event_data:
            if match_element is None:
                if not self.allow_missing_values_flag:
                    msg = 'Empty entry detected in event_data.' \
                        'Please fill entry or set parameter allow_missing_values_flag to true.'
                    logging.getLogger(DEBUG_LOG_NAME).error(msg)
                    raise Exception(msg)
                match_value_list.append(None)
            else:
                match_value_list.append(bytes(match_element, 'utf-8'))

        match_value_tuple = tuple(match_value_list)
        if match_value_tuple not in self.known_values_set:
            self.known_values_set.add(match_value_tuple)
            self.log_learned_path_value_combos += 1
            self.log_new_learned_values.append(match_value_tuple)
        return f"Added values [{', '.join(event_data)}] of paths [{', '.join(self.target_path_list)}] to the persistence."

    def log_statistics(self, component_name):
        """
        Log statistics of an AtomHandler. Override this method for more sophisticated statistics output of the AtomHandler.
        @param component_name the name of the component which is printed in the log line.
        """
        if AminerConfig.STAT_LEVEL == 1:
            logging.getLogger(STAT_LOG_NAME).info(
                f"'{component_name}' processed {self.log_success} out of {self.log_total} log atoms successfully and learned "
                f"{self.log_learned_path_value_combos} new value combinations in the last 60 minutes.")
        elif AminerConfig.STAT_LEVEL == 2:
            logging.getLogger(STAT_LOG_NAME).info(
                f"'{component_name}' processed {self.log_success} out of {self.log_total} log atoms successfully and learned "
                f"{self.log_learned_path_value_combos} new value combinations in the last 60 minutes. Following new value combinations "
                f"were learned: {self.log_new_learned_values}")
        self.log_success = 0
        self.log_total = 0
        self.log_learned_path_value_combos = 0
        self.log_new_learned_values = []



import unittest
from aminer.parsing.RepeatedElementDataModelElement import RepeatedElementDataModelElement
from aminer.parsing.MatchContext import MatchContext
from aminer.parsing.MatchElement import MatchElement
from unit.TestBase import TestBase, DummyMatchContext, DummyFixedDataModelElement


class RepeatedElementDataModelElementTest(TestBase):
    """Unittests for the RepeatedElementDataModelElement."""

    id_ = "repeated"
    path = "path"
    fixed_id = "fixed"
    fixed_data = b"fixed data "

    def test1get_match_element_valid_match(self):
        """Parse matching substring from MatchContext and check if the MatchContext was updated with all characters."""
        fixed_dme = DummyFixedDataModelElement(self.fixed_id, self.fixed_data)
        repeated_dme = RepeatedElementDataModelElement(self.id_, DummyFixedDataModelElement(self.fixed_id, self.fixed_data))
        data = b"fixed data string."
        value = b"fixed data "
        match_context = DummyMatchContext(data)
        match_element = repeated_dme.get_match_element(self.path, match_context)
        self.compare_match_results(data, match_element, match_context, self.id_, self.path, value, value, [
            fixed_dme.get_match_element("%s/%s/0" % (self.path, self.id_), DummyMatchContext(data))])

        data = b"fixed data fixed data fixed data fixed data "
        match_context = DummyMatchContext(data)
        match_element = repeated_dme.get_match_element(self.path, match_context)
        self.compare_match_results(data, match_element, match_context, self.id_, self.path, data, data, [
            fixed_dme.get_match_element("%s/%s/0" % (self.path, self.id_), DummyMatchContext(data)),
            fixed_dme.get_match_element("%s/%s/1" % (self.path, self.id_), DummyMatchContext(data)),
            fixed_dme.get_match_element("%s/%s/2" % (self.path, self.id_), DummyMatchContext(data)),
            fixed_dme.get_match_element("%s/%s/3" % (self.path, self.id_), DummyMatchContext(data))
        ])

        data = b"fixed data fixed data \nhere is some other string.\nfixed data fixed data "
        value = b"fixed data fixed data "
        match_context = DummyMatchContext(data)
        match_element = repeated_dme.get_match_element(self.path, match_context)
        self.compare_match_results(data, match_element, match_context, self.id_, self.path, value, value, [
            fixed_dme.get_match_element("%s/%s/0" % (self.path, self.id_), DummyMatchContext(data)),
            fixed_dme.get_match_element("%s/%s/1" % (self.path, self.id_), DummyMatchContext(data))
        ])

    def test2get_match_element_min_max_repeats(self):
        """This test case verifies the functionality of setting the minimal and maximal repeats."""
        fixed_dme = DummyFixedDataModelElement(self.fixed_id, self.fixed_data)
        repeated_dme = RepeatedElementDataModelElement(self.id_, fixed_dme, min_repeat=2, max_repeat=5)
        same_min_max_repeat_dme = RepeatedElementDataModelElement(self.id_, fixed_dme, min_repeat=3, max_repeat=3)
        data = b"other data"
        match_context = DummyMatchContext(data)
        match_element = repeated_dme.get_match_element(self.path, match_context)
        self.compare_no_match_results(data, match_element, match_context)
        match_context = DummyMatchContext(data)
        match_element = same_min_max_repeat_dme.get_match_element(self.path, match_context)
        self.compare_no_match_results(data, match_element, match_context)

        data = b"fixed data "
        match_context = DummyMatchContext(data)
        match_element = repeated_dme.get_match_element(self.path, match_context)
        self.compare_no_match_results(data, match_element, match_context)

        match_context = DummyMatchContext(data)
        match_element = same_min_max_repeat_dme.get_match_element(self.path, match_context)
        self.compare_no_match_results(data, match_element, match_context)

        data = b"fixed data fixed data "
        match_context = DummyMatchContext(data)
        match_element = repeated_dme.get_match_element(self.path, match_context)
        self.compare_match_results(data, match_element, match_context, self.id_, self.path, data, data, [
            fixed_dme.get_match_element("%s/%s/0" % (self.path, self.id_), DummyMatchContext(data)),
            fixed_dme.get_match_element("%s/%s/1" % (self.path, self.id_), DummyMatchContext(data))])

        match_context = DummyMatchContext(data)
        match_element = same_min_max_repeat_dme.get_match_element(self.path, match_context)
        self.compare_no_match_results(data, match_element, match_context)

        data = b"fixed data fixed data fixed data "
        match_context = DummyMatchContext(data)
        match_element = repeated_dme.get_match_element(self.path, match_context)
        self.compare_match_results(data, match_element, match_context, self.id_, self.path, data, data, [
            fixed_dme.get_match_element("%s/%s/0" % (self.path, self.id_), DummyMatchContext(data)),
            fixed_dme.get_match_element("%s/%s/1" % (self.path, self.id_), DummyMatchContext(data)),
            fixed_dme.get_match_element("%s/%s/2" % (self.path, self.id_), DummyMatchContext(data))])

        match_context = DummyMatchContext(data)
        match_element = same_min_max_repeat_dme.get_match_element(self.path, match_context)
        self.compare_match_results(data, match_element, match_context, self.id_, self.path, data, data, [
            fixed_dme.get_match_element("%s/%s/0" % (self.path, self.id_), DummyMatchContext(data)),
            fixed_dme.get_match_element("%s/%s/1" % (self.path, self.id_), DummyMatchContext(data)),
            fixed_dme.get_match_element("%s/%s/2" % (self.path, self.id_), DummyMatchContext(data))])

        data = b"fixed data fixed data fixed data fixed data "
        match_context = DummyMatchContext(data)
        match_element = repeated_dme.get_match_element(self.path, match_context)
        self.compare_match_results(data, match_element, match_context, self.id_, self.path, data, data, [
            fixed_dme.get_match_element("%s/%s/0" % (self.path, self.id_), DummyMatchContext(data)),
            fixed_dme.get_match_element("%s/%s/1" % (self.path, self.id_), DummyMatchContext(data)),
            fixed_dme.get_match_element("%s/%s/2" % (self.path, self.id_), DummyMatchContext(data)),
            fixed_dme.get_match_element("%s/%s/3" % (self.path, self.id_), DummyMatchContext(data))])

        match_context = DummyMatchContext(data)
        match_element = same_min_max_repeat_dme.get_match_element(self.path, match_context)
        self.compare_no_match_results(data, match_element, match_context)

        data = b"fixed data fixed data fixed data fixed data fixed data "
        match_context = DummyMatchContext(data)
        match_element = repeated_dme.get_match_element(self.path, match_context)
        self.compare_match_results(data, match_element, match_context, self.id_, self.path, data, data, [
            fixed_dme.get_match_element("%s/%s/0" % (self.path, self.id_), DummyMatchContext(data)),
            fixed_dme.get_match_element("%s/%s/1" % (self.path, self.id_), DummyMatchContext(data)),
            fixed_dme.get_match_element("%s/%s/2" % (self.path, self.id_), DummyMatchContext(data)),
            fixed_dme.get_match_element("%s/%s/3" % (self.path, self.id_), DummyMatchContext(data)),
            fixed_dme.get_match_element("%s/%s/4" % (self.path, self.id_), DummyMatchContext(data))])

        match_context = DummyMatchContext(data)
        match_element = same_min_max_repeat_dme.get_match_element(self.path, match_context)
        self.compare_no_match_results(data, match_element, match_context)

        data = b"fixed data fixed data fixed data fixed data fixed data fixed data "
        match_context = DummyMatchContext(data)
        match_element = repeated_dme.get_match_element(self.path, match_context)
        self.compare_no_match_results(data, match_element, match_context)

        match_context = DummyMatchContext(data)
        match_element = same_min_max_repeat_dme.get_match_element(self.path, match_context)
        self.compare_no_match_results(data, match_element, match_context)

    def test3element_id_input_validation(self):
        """Check if element_id is validated."""
        fixed_dme = DummyFixedDataModelElement(self.fixed_id, self.fixed_data)
        self.assertRaises(ValueError, RepeatedElementDataModelElement, "", fixed_dme)
        self.assertRaises(TypeError, RepeatedElementDataModelElement, None, fixed_dme)
        self.assertRaises(TypeError, RepeatedElementDataModelElement, b"path", fixed_dme)
        self.assertRaises(TypeError, RepeatedElementDataModelElement, True, fixed_dme)
        self.assertRaises(TypeError, RepeatedElementDataModelElement, 123, fixed_dme)
        self.assertRaises(TypeError, RepeatedElementDataModelElement, 123.22, fixed_dme)
        self.assertRaises(TypeError, RepeatedElementDataModelElement, {"id": "path"}, fixed_dme)
        self.assertRaises(TypeError, RepeatedElementDataModelElement, ["path"], fixed_dme)
        self.assertRaises(TypeError, RepeatedElementDataModelElement, [], fixed_dme)
        self.assertRaises(TypeError, RepeatedElementDataModelElement, (), fixed_dme)
        self.assertRaises(TypeError, RepeatedElementDataModelElement, set(), fixed_dme)

    def test4repeated_element_input_validation(self):
        """Check if repeated_element is validated."""
        self.assertRaises(TypeError, RepeatedElementDataModelElement, self.id_, "string")
        self.assertRaises(TypeError, RepeatedElementDataModelElement, self.id_, None)
        self.assertRaises(TypeError, RepeatedElementDataModelElement, self.id_, b"path")
        self.assertRaises(TypeError, RepeatedElementDataModelElement, self.id_, True)
        self.assertRaises(TypeError, RepeatedElementDataModelElement, self.id_, 123)
        self.assertRaises(TypeError, RepeatedElementDataModelElement, self.id_, 123.22)
        self.assertRaises(TypeError, RepeatedElementDataModelElement, self.id_, {"id": "path"})
        self.assertRaises(TypeError, RepeatedElementDataModelElement, self.id_, ["path"])
        self.assertRaises(TypeError, RepeatedElementDataModelElement, self.id_, [])
        self.assertRaises(TypeError, RepeatedElementDataModelElement, self.id_, ())
        self.assertRaises(TypeError, RepeatedElementDataModelElement, self.id_, set())

    def test5min_repeat_input_validation(self):
        """Check if min_repeat is validated."""
        fixed_dme = DummyFixedDataModelElement(self.fixed_id, self.fixed_data)
        self.assertRaises(TypeError, RepeatedElementDataModelElement, self.id_, fixed_dme, min_repeat="string")
        self.assertRaises(TypeError, RepeatedElementDataModelElement, self.id_, fixed_dme, min_repeat=None)
        self.assertRaises(TypeError, RepeatedElementDataModelElement, self.id_, fixed_dme, min_repeat=b"path")
        self.assertRaises(TypeError, RepeatedElementDataModelElement, self.id_, fixed_dme, min_repeat=True)
        self.assertRaises(ValueError, RepeatedElementDataModelElement, self.id_, fixed_dme, min_repeat=-1)
        self.assertRaises(TypeError, RepeatedElementDataModelElement, self.id_, fixed_dme, min_repeat=123.22)
        self.assertRaises(TypeError, RepeatedElementDataModelElement, self.id_, fixed_dme, min_repeat={"id": "path"})
        self.assertRaises(TypeError, RepeatedElementDataModelElement, self.id_, fixed_dme, min_repeat=["path"])
        self.assertRaises(TypeError, RepeatedElementDataModelElement, self.id_, fixed_dme, min_repeat=[])
        self.assertRaises(TypeError, RepeatedElementDataModelElement, self.id_, fixed_dme, min_repeat=())
        self.assertRaises(TypeError, RepeatedElementDataModelElement, self.id_, fixed_dme, min_repeat=set())

    def test6max_repeat_input_validation(self):
        """Check if max_repeat is validated."""
        fixed_dme = DummyFixedDataModelElement(self.fixed_id, self.fixed_data)
        self.assertRaises(TypeError, RepeatedElementDataModelElement, self.id_, fixed_dme, max_repeat="string")
        self.assertRaises(TypeError, RepeatedElementDataModelElement, self.id_, fixed_dme, max_repeat=None)
        self.assertRaises(TypeError, RepeatedElementDataModelElement, self.id_, fixed_dme, max_repeat=b"path")
        self.assertRaises(TypeError, RepeatedElementDataModelElement, self.id_, fixed_dme, max_repeat=True)
        self.assertRaises(ValueError, RepeatedElementDataModelElement, self.id_, fixed_dme, max_repeat=0)
        self.assertRaises(ValueError, RepeatedElementDataModelElement, self.id_, fixed_dme, max_repeat=10, min_repeat=11)
        self.assertRaises(TypeError, RepeatedElementDataModelElement, self.id_, fixed_dme, max_repeat=123.22)
        self.assertRaises(TypeError, RepeatedElementDataModelElement, self.id_, fixed_dme, max_repeat={"id": "path"})
        self.assertRaises(TypeError, RepeatedElementDataModelElement, self.id_, fixed_dme, max_repeat=["path"])
        self.assertRaises(TypeError, RepeatedElementDataModelElement, self.id_, fixed_dme, max_repeat=[])
        self.assertRaises(TypeError, RepeatedElementDataModelElement, self.id_, fixed_dme, max_repeat=())
        self.assertRaises(TypeError, RepeatedElementDataModelElement, self.id_, fixed_dme, max_repeat=set())

    def test7get_match_element_match_context_input_validation(self):
        """Check if an exception is raised, when other classes than MatchContext are used in get_match_element."""
        model_element = RepeatedElementDataModelElement(self.id_, DummyFixedDataModelElement(self.fixed_id, self.fixed_data))
        data = b"fixed data"
        model_element.get_match_element(self.path, DummyMatchContext(data))
        model_element.get_match_element(self.path, MatchContext(data))

        self.assertRaises(AttributeError, model_element.get_match_element, self.path, MatchElement(None, data, None, None))
        self.assertRaises(AttributeError, model_element.get_match_element, self.path, data)
        self.assertRaises(AttributeError, model_element.get_match_element, self.path, data.decode())
        self.assertRaises(AttributeError, model_element.get_match_element, self.path, 123)
        self.assertRaises(AttributeError, model_element.get_match_element, self.path, 123.22)
        self.assertRaises(AttributeError, model_element.get_match_element, self.path, True)
        self.assertRaises(AttributeError, model_element.get_match_element, self.path, None)
        self.assertRaises(AttributeError, model_element.get_match_element, self.path, [])
        self.assertRaises(AttributeError, model_element.get_match_element, self.path, {"key": MatchContext(data)})
        self.assertRaises(AttributeError, model_element.get_match_element, self.path, set())
        self.assertRaises(AttributeError, model_element.get_match_element, self.path, ())
        self.assertRaises(AttributeError, model_element.get_match_element, self.path, model_element)


if __name__ == "__main__":
    unittest.main()



"""This module defines a generated parser model."""

from aminer.parsing.DateTimeModelElement import DateTimeModelElement
from aminer.parsing.DecimalIntegerValueModelElement import DecimalIntegerValueModelElement
from aminer.parsing.DelimitedDataModelElement import DelimitedDataModelElement
from aminer.parsing.FixedDataModelElement import FixedDataModelElement
from aminer.parsing.IpAddressDataModelElement import IpAddressDataModelElement
from aminer.parsing.SequenceModelElement import SequenceModelElement


def get_model():
    """Return a model to parse Suricata Fast logs from the AIT-LDS."""
    model = SequenceModelElement("model", [
        DateTimeModelElement("time", b"%m/%d/%Y-%H:%M:%S.%f"),
        FixedDataModelElement("brack_str1", b"  [**] ["),
        DecimalIntegerValueModelElement("id1"),
        FixedDataModelElement("sep1", b":"),
        DecimalIntegerValueModelElement("id2"),
        FixedDataModelElement("sep2", b":"),
        DecimalIntegerValueModelElement("id3"),
        FixedDataModelElement("sep3", b"] "),
        DelimitedDataModelElement("message", b" [**] "),
        FixedDataModelElement("classification_str", b" [**] [Classification: "),
        DelimitedDataModelElement("classification", b"]"),
        FixedDataModelElement("priority_str", b"] [Priority: "),
        DecimalIntegerValueModelElement("priority"),
        FixedDataModelElement("brack_str1", b"] {"),
        DelimitedDataModelElement("conn", b"}"),
        FixedDataModelElement("brack_str2", b"} "),
        IpAddressDataModelElement("src_ip"),
        FixedDataModelElement("colon", b":"),
        DecimalIntegerValueModelElement("src_port"),
        FixedDataModelElement("arrow_str", b" -> "),
        IpAddressDataModelElement("dst_ip"),
        FixedDataModelElement("colon", b":"),
        DecimalIntegerValueModelElement("dst_port"),
        ])
    return model



import unittest
import os
import shutil
import logging
import sys
import errno
import inspect
from aminer.AminerConfig import KEY_LOG_DIR, DEFAULT_LOG_DIR, KEY_PERSISTENCE_DIR, DEFAULT_PERSISTENCE_DIR, DEBUG_LOG_NAME,\
    KEY_REMOTE_CONTROL_LOG_FILE, KEY_STAT_LOG_FILE, KEY_DEBUG_LOG_FILE, REMOTE_CONTROL_LOG_NAME, DEFAULT_REMOTE_CONTROL_LOG_FILE,\
    STAT_LOG_NAME, DEFAULT_STAT_LOG_FILE, DEBUG_LEVEL, load_config, build_persistence_file_name, DEFAULT_DEBUG_LOG_FILE
from aminer.AnalysisChild import AnalysisContext
from aminer.events.StreamPrinterEventHandler import StreamPrinterEventHandler
from aminer.parsing.ModelElementInterface import ModelElementInterface
from aminer.parsing.MatchElement import MatchElement
from aminer.util import PersistenceUtil
from aminer.util import SecureOSFunctions
from _io import StringIO


def initialize_loggers(aminer_config, aminer_user_id, aminer_grp_id):
    """Initialize all loggers."""
    datefmt = '%d/%b/%Y:%H:%M:%S %z'

    log_dir = aminer_config.config_properties.get(KEY_LOG_DIR, DEFAULT_LOG_DIR)
    if log_dir == DEFAULT_LOG_DIR:
        try:
            if not os.path.isdir(log_dir):
                persistence_dir_path = aminer_config.config_properties.get(KEY_PERSISTENCE_DIR, DEFAULT_PERSISTENCE_DIR)
                persistence_dir_fd = SecureOSFunctions.secure_open_base_directory(persistence_dir_path)
                if SecureOSFunctions.base_dir_path == DEFAULT_PERSISTENCE_DIR:
                    relative_path_log_dir = os.path.split(DEFAULT_LOG_DIR)[1]
                    os.mkdir(relative_path_log_dir, dir_fd=persistence_dir_fd)
                    os.chown(relative_path_log_dir, aminer_user_id, aminer_grp_id, dir_fd=persistence_dir_fd, follow_symlinks=False)
        except OSError as e:
            if e.errno != errno.EEXIST:
                msg = 'Unable to create log-directory: %s' % log_dir
            else:
                msg = e
            logging.getLogger(DEBUG_LOG_NAME).error(msg.strip('\n'))
            print(msg, file=sys.stderr)

    tmp_value = aminer_config.config_properties.get(KEY_REMOTE_CONTROL_LOG_FILE)
    if tmp_value is not None and b'/' in tmp_value:
        print('%s attribute must not contain a full directory path, but only the filename.' % KEY_REMOTE_CONTROL_LOG_FILE,
              file=sys.stderr)
        sys.exit(1)
    tmp_value = aminer_config.config_properties.get(KEY_STAT_LOG_FILE)
    if tmp_value is not None and b'/' in tmp_value:
        print('%s attribute must not contain a full directory path, but only the filename.' % KEY_STAT_LOG_FILE, file=sys.stderr)
        sys.exit(1)
    tmp_value = aminer_config.config_properties.get(KEY_DEBUG_LOG_FILE)
    if tmp_value is not None and b'/' in tmp_value:
        print('%s attribute must not contain a full directory path, but only the filename.' % KEY_DEBUG_LOG_FILE, file=sys.stderr)
        sys.exit(1)

    log_dir_fd = SecureOSFunctions.secure_open_log_directory(log_dir, os.O_RDONLY | os.O_DIRECTORY | os.O_PATH)
    rc_logger = logging.getLogger(REMOTE_CONTROL_LOG_NAME)
    rc_logger.setLevel(logging.DEBUG)
    remote_control_log_file = aminer_config.config_properties.get(
        KEY_REMOTE_CONTROL_LOG_FILE, os.path.join(log_dir, DEFAULT_REMOTE_CONTROL_LOG_FILE))
    if not remote_control_log_file.startswith(log_dir):
        remote_control_log_file = os.path.join(log_dir, remote_control_log_file)
    try:
        rc_file_handler = logging.FileHandler(remote_control_log_file)
        os.chown(remote_control_log_file, aminer_user_id, aminer_grp_id, dir_fd=log_dir_fd, follow_symlinks=False)
    except OSError as e:
        print('Could not create or open %s: %s. Stopping..' % (remote_control_log_file, e), file=sys.stderr)
        sys.exit(1)
    rc_file_handler.setFormatter(logging.Formatter(fmt='%(asctime)s %(levelname)s %(message)s', datefmt=datefmt))
    rc_logger.addHandler(rc_file_handler)
    logging.addLevelName(15, "REMOTECONTROL")

    stat_logger = logging.getLogger(STAT_LOG_NAME)
    stat_logger.setLevel(logging.INFO)
    stat_log_file = aminer_config.config_properties.get(KEY_STAT_LOG_FILE, os.path.join(log_dir, DEFAULT_STAT_LOG_FILE))
    if not stat_log_file.startswith(log_dir):
        stat_log_file = os.path.join(log_dir, stat_log_file)
    try:
        stat_file_handler = logging.FileHandler(stat_log_file)
        os.chown(stat_log_file, aminer_user_id, aminer_grp_id, dir_fd=log_dir_fd, follow_symlinks=False)
    except OSError as e:
        print('Could not create or open %s: %s. Stopping..' % (stat_log_file, e), file=sys.stderr)
        sys.exit(1)
    stat_file_handler.setFormatter(logging.Formatter(fmt='%(asctime)s %(message)s', datefmt=datefmt))
    stat_logger.addHandler(stat_file_handler)

    debug_logger = logging.getLogger(DEBUG_LOG_NAME)
    if DEBUG_LEVEL == 0:
        debug_logger.setLevel(logging.ERROR)
    elif DEBUG_LEVEL == 1:
        debug_logger.setLevel(logging.INFO)
    else:
        debug_logger.setLevel(logging.DEBUG)
    debug_log_file = aminer_config.config_properties.get(
        KEY_DEBUG_LOG_FILE, os.path.join(log_dir, DEFAULT_DEBUG_LOG_FILE))
    if not debug_log_file.startswith(log_dir):
        debug_log_file = os.path.join(log_dir, debug_log_file)
    try:
        debug_file_handler = logging.FileHandler(debug_log_file)
        os.chown(debug_log_file, aminer_user_id, aminer_grp_id, dir_fd=log_dir_fd, follow_symlinks=False)
    except OSError as e:
        print('Could not create or open %s: %s. Stopping..' % (debug_log_file, e), file=sys.stderr)
        sys.exit(1)
    debug_file_handler.setFormatter(logging.Formatter(fmt='%(asctime)s %(levelname)s %(message)s', datefmt=datefmt))
    debug_logger.addHandler(debug_file_handler)


# skipcq: PTC-W0046
class TestBase(unittest.TestCase):
    """This is the base class for all unittests."""

    def get_config_file_path(self):
        """
        Get the module name to choose the right config file for parallel execution.
        Example: logdata-anomaly-miner/aecid-testsuite/unit/analysis/AtomFiltersTest.py - we want to know the directory analysis.
        """
        return os.getcwd()+'/unit/data/parallel_configs/%s_config.py' % inspect.getmodule(self).__file__.split("unit/")[1].split("/")[0]

    def setUp(self):
        """Set up all needed variables and remove persisted data."""
        PersistenceUtil.persistable_components = []
        self.aminer_config = load_config(self.get_config_file_path())
        self.analysis_context = AnalysisContext(self.aminer_config)
        self.output_stream = StringIO()
        self.stream_printer_event_handler = StreamPrinterEventHandler(self.analysis_context, self.output_stream)
        persistence_dir_name = build_persistence_file_name(self.aminer_config)
        if os.path.exists(persistence_dir_name):
            shutil.rmtree(persistence_dir_name)
        if not os.path.exists(persistence_dir_name):
            os.makedirs(persistence_dir_name)
        initialize_loggers(self.aminer_config, os.getuid(), os.getgid())
        if isinstance(persistence_dir_name, str):
            persistence_dir_name = persistence_dir_name.encode()
        SecureOSFunctions.secure_open_base_directory(persistence_dir_name, os.O_RDONLY | os.O_DIRECTORY | os.O_PATH)
        PersistenceUtil.SKIP_PERSISTENCE_ID_WARNING = True

    def tearDown(self):
        """Delete all persisted data after the tests."""
        self.aminer_config = load_config(self.get_config_file_path())
        persistence_file_name = build_persistence_file_name(self.aminer_config)
        if os.path.exists(persistence_file_name):
            shutil.rmtree(persistence_file_name)
        if not os.path.exists(persistence_file_name):
            os.makedirs(persistence_file_name)
        SecureOSFunctions.close_base_directory()

    def reset_output_stream(self):
        """Reset the output stream."""
        self.output_stream.seek(0)
        self.output_stream.truncate(0)

    def compare_match_results(self, data, match_element, match_context, id_, path, match_string, match_object, children):
        """Compare the results of get_match_element() if match_element is not None."""
        self.assertEqual(match_element.path, "%s/%s" % (path, id_))
        self.assertEqual(match_element.match_string, match_string)
        self.assertEqual(match_element.match_object, match_object)
        if children is None:
            self.assertIsNone(match_element.children, children)
        else:
            self.assertEqual(len(children), len(match_element.children))
            for i, child in enumerate(children):
                self.assertEqual(match_element.children[i].path, child.path)
                self.assertEqual(match_element.children[i].match_string, child.match_string)
                self.assertEqual(match_element.children[i].match_object, child.match_object)
                self.assertIsNone(match_element.children[i].children, children)
        self.assertEqual(match_context.match_string, match_string)
        self.assertEqual(match_context.match_data, data[len(match_string):])

    def compare_no_match_results(self, data, match_element, match_context):
        """Compare the results of get_match_element() if match_element is not None."""
        self.assertIsNone(match_element, None)
        self.assertEqual(match_context.match_data, data)


class DummyMatchContext:
    """Dummy class for MatchContext."""

    def __init__(self, match_data: bytes):
        """Initiate the Dummy class."""
        self.match_data = match_data
        self.match_string = b''

    def update(self, match_string: bytes):
        """Update the data."""
        self.match_data = self.match_data[len(match_string):]
        self.match_string += match_string


class DummyFixedDataModelElement(ModelElementInterface):
    """Dummy class for fixed string ModelElements."""

    def __init__(self, element_id: str, data: bytes):
        self.element_id = element_id
        self.data = data

    def get_id(self):
        """Get the element ID."""
        return self.element_id

    def get_child_elements(self):  # skipcq: PYL-R0201
        """
        Get all possible child model elements of this element.
        @return None as there are no children of this element.
        """
        return None

    def get_match_element(self, path: str, match_context):
        """@return None when there is no match, MatchElement otherwise."""
        if not match_context.match_data.startswith(self.data):
            return None
        match_context.update(self.data)
        return MatchElement("%s/%s" % (path, self.element_id), self.data, self.data, None)


class DummyFirstMatchModelElement(ModelElementInterface):
    """This class defines a model element to return the match from the the first matching child model within a given list."""

    def __init__(self, element_id, children):
        self.element_id = element_id
        self.children = children
        if (children is None) or (None in children):
            msg = 'Invalid children list'
            logging.getLogger(DEBUG_LOG_NAME).error(msg)
            raise Exception(msg)

    def get_id(self):
        """Get the element ID."""
        return self.element_id

    def get_child_elements(self):
        """Get all possible child model elements of this element."""
        return self.children

    def get_match_element(self, path, match_context):
        """@return None when there is no match, MatchElement otherwise."""
        current_path = "%s/%s" % (path, self.element_id)

        match_data = match_context.match_data
        for child_element in self.children:
            child_match = child_element.get_match_element(current_path, match_context)
            if child_match is not None:
                return child_match
            match_context.match_data = match_data
        return None


if __name__ == "__main__":
    unittest.main()



"""
This module defines the match context.

This program is free software: you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.
This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
You should have received a copy of the GNU General Public License along with
this program. If not, see <http://www.gnu.org/licenses/>.
"""
import logging
from typing import Union
from aminer.AminerConfig import DEBUG_LOG_NAME
from aminer import AminerConfig


class MatchContext:
    """
    This class allows storage of data relevant during the matching process, e.g. the root node and the remaining unmatched data.
    Then searching for non-atomic matches, e.g. sequences, the context might be modified by model subelements, even if the main model
    element will not return a match. In that case, those non-atomic model elements have to care to restore the context before returning.
    """

    def __init__(self, match_data: bytes):
        """
        Create a MatchContext with the full unmatched string data.
        @param match_data the data that will be tested by the next model element.
        """
        if not isinstance(match_data, bytes):
            msg = "match_data has to be of the type bytes."
            logging.getLogger(DEBUG_LOG_NAME).error(msg)
            raise TypeError(msg)
        self.match_data = match_data

    def update(self, match_string: bytes):
        """
        Update the match context by removing the given matched string data from the context data still to be matched.
        This method does not check, if the removed data is the same as the trailing match data for performance reasons. This is done
        only in the DebugMatchContext class.
        """
        self.match_data = self.match_data[len(match_string):]


class DebugMatchContext(MatchContext):
    """This class defines a slower MatchContext for debugging purposes."""

    def __init__(self, match_data: bytes):
        self.debug_info = ""
        self.last_match_data: Union[None, bytes] = None
        self.shortest_unmatched_data = match_data
        super(DebugMatchContext, self).__init__(match_data)

    def update(self, match_string: bytes):
        """Update the context and store debugging information."""
        if not isinstance(match_string, bytes):
            msg = "match_string has to be of the type bytes."
            logging.getLogger(DEBUG_LOG_NAME).error(msg)
            raise TypeError(msg)
        if len(match_string) < 1:
            return
        try:
            match_data = self.match_data.decode(AminerConfig.ENCODING)
            m_string = match_string.decode(AminerConfig.ENCODING)
        except UnicodeError:
            match_data = repr(self.match_data)
            m_string = repr(match_string)
        if self.last_match_data != self.match_data:
            self.last_match_data = self.match_data
            if self.debug_info != "":
                self.debug_info += "  "
            self.debug_info += f'Starting match update on "{match_data}"\n'
        if not self.match_data.startswith(match_string):
            self.debug_info += f'  Current data {match_data} does not start with "{m_string}"\n'
            msg = "Illegal state"
            logging.getLogger(DEBUG_LOG_NAME).error(msg)
            raise ValueError(msg)
        self.match_data = self.match_data[len(match_string):]
        self.last_match_data = self.match_data
        if (self.shortest_unmatched_data is None) or (len(self.match_data) < len(self.shortest_unmatched_data)):
            self.shortest_unmatched_data = self.match_data
        self.debug_info += f'  Removed: "{m_string}", remaining {len(self.match_data)} bytes\n'

    def get_debug_info(self):
        """Get the current debugging information and reset it."""
        while self.debug_info.find("\n\n") != -1:
            self.debug_info = self.debug_info.replace("\n\n", "\n")
        result = self.debug_info
        self.debug_info = ""
        try:
            data = self.shortest_unmatched_data.decode(AminerConfig.ENCODING)
        except UnicodeError:
            data = repr(self.shortest_unmatched_data)
        result += f'  Shortest unmatched data: "{data}"\n'
        return result

    def get_shortest_unmatched_data(self):
        """
        Get the shortest match_data found while updating the internal state.
        This is useful to find out where the parsing process has terminated.
        """
        return self.shortest_unmatched_data



"""
This module defines a PCA-detector for event and value counts.
The component detects anomalies by creating an Event-Count-Matrix for given
time-windows to calculate an anomaly score for new time windows afterwards by
using the reconstruction error from the inverse-transformation with restricted
components of the Principal-Component-Analysis (PCA).

This program is free software: you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.
This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
You should have received a copy of the GNU General Public License along with
this program. If not, see <http://www.gnu.org/licenses/>.
"""
import copy
import numpy as np
import logging
import os
import time
from aminer import AminerConfig
from aminer.AminerConfig import DEBUG_LOG_NAME, STAT_LEVEL, STAT_LOG_NAME, CONFIG_KEY_LOG_LINE_PREFIX, DEFAULT_LOG_LINE_PREFIX,\
    KEY_PERSISTENCE_PERIOD, DEFAULT_PERSISTENCE_PERIOD
from aminer.AnalysisChild import AnalysisContext
from aminer.util import PersistenceUtil
from aminer.input.InputInterfaces import AtomHandlerInterface
from aminer.util.TimeTriggeredComponentInterface import TimeTriggeredComponentInterface


class PCADetector(AtomHandlerInterface, TimeTriggeredComponentInterface):
    """This class creates events if event or value occurrence counts are outliers in PCA space."""

    time_trigger_class = AnalysisContext.TIME_TRIGGER_CLASS_REALTIME

    def __init__(self, aminer_config, target_path_list, anomaly_event_handlers, window_size, min_anomaly_score, min_variance, num_windows,
                 persistence_id='Default', learn_mode=False, output_logline=True, ignore_list=None, constraint_list=None,
                 stop_learning_time=None, stop_learning_no_anomaly_time=None):
        """
        Initialize the detector. This will also trigger reading or creation of persistence storage location.
        @param aminer_config configuration from analysis_context.
        @param target_path_list parser paths of values to be analyzed. Multiple paths mean that values are analyzed as separate
               dimensions. When no paths are specified, the events given by the full path list are analyzed (one dimension).
        @param anomaly_event_handlers for handling events, e.g., print events to stdout.
        @param window_size the length of the time window for counting in seconds.
        @param min_anomaly_score the minimum computed outlier score for reporting anomalies. Scores are scaled by training data, i.e.,
               reasonable minimum scores are >1 to detect outliers with respect to currently trained PCA matrix.
        @param min_variance the minimum variance covered by the principal components in range [0, 1].
        @param num_windows the number of time windows in the sliding window approach. Total covered time span = window_size * num_windows.
        @param persistence_id name of persistence file.
        @param learn_mode specifies whether new count measurements are added to the PCA count matrix.
        @param output_logline specifies whether the full parsed log atom should be provided in the output.
        @param ignore_list list of paths that are not considered for analysis, i.e., events that contain one of these paths are
               omitted. The default value is [] as None is not iterable.
        @param constraint_list list of paths that have to be present in the log atom to be analyzed.
        @param stop_learning_time switch the learn_mode to False after the time.
        @param stop_learning_no_anomaly_time switch the learn_mode to False after no anomaly was detected for that time.
        """
        # avoid "defined outside init" issue
        self.learn_mode, self.stop_learning_timestamp, self.next_persist_time, self.log_success, self.log_total = [None]*5
        super().__init__(
            mutable_default_args=["ignore_list", "constraint_list"], aminer_config=aminer_config, target_path_list=target_path_list,
            anomaly_event_handlers=anomaly_event_handlers, window_size=window_size, min_anomaly_score=min_anomaly_score,
            min_variance=min_variance, num_windows=num_windows, persistence_id=persistence_id, learn_mode=learn_mode,
            output_logline=output_logline, ignore_list=ignore_list, constraint_list=constraint_list, stop_learning_time=stop_learning_time,
            stop_learning_no_anomaly_time=stop_learning_no_anomaly_time
        )
        # skipcq: PYL-W0511
        # ToDo: an exception should be thrown instead of this check.
        if num_windows < 3:
            logging.getLogger(AminerConfig.DEBUG_LOG_NAME).warning('num_windows must be >= 3!')
            self.num_windows = 3
        else:
            self.num_windows = num_windows
        self.first_log = True
        self.start_time = 0
        self.event_count_matrix = []
        self.feature_list = []
        self.ecm = None
        self.log_windows = 0
        self.pca_ecm = None
        self.eigen_vectors = None
        # number of components (n_comp): how many components should be used for reconstruction
        self.n_comp = None
        # Calculate Anomaly-Score (Reconstruction Error) for the whole dataset
        self.loss = None

        self.persistence_file_name = AminerConfig.build_persistence_file_name(aminer_config, self.__class__.__name__, persistence_id)
        PersistenceUtil.add_persistable_component(self)
        persistence_data = PersistenceUtil.load_json(self.persistence_file_name)

        if persistence_data is not None:
            self.event_count_matrix = list(persistence_data)
            self.compute_pca()
            # Copy feature list into event count vector and reset counts of each feature
            self.event_count_vector = copy.deepcopy(self.event_count_matrix[0])
            self.reset_event_count_vector()
        else:
            if self.target_path_list is None or len(self.target_path_list) == 0:
                # Only one dimension when events are used instead of values; use empty string as placeholder
                self.event_count_vector = {'': {}}
            else:
                self.event_count_vector = {}

    def receive_atom(self, log_atom):
        """Receive parsed atom and the information about the parser match."""
        parser_match = log_atom.parser_match
        self.log_total += 1
        if self.learn_mode is True and self.stop_learning_timestamp is not None and \
                self.stop_learning_timestamp < log_atom.atom_time:
            logging.getLogger(DEBUG_LOG_NAME).info(f"Stopping learning in the {self.__class__.__name__}.")
            self.learn_mode = False

        # Skip paths from ignore list.
        for ignore_path in self.ignore_list:
            if ignore_path in parser_match.get_match_dictionary().keys():
                return

        # get the timestamp of the first log to start the time-window-process (flag)
        if self.first_log:
            self.start_time = log_atom.get_timestamp()
            self.first_log = False

        current_time = log_atom.get_timestamp()
        while current_time >= (self.start_time + self.window_size):
            # PCA computation only possible when at least 3 vectors are present
            if len(self.event_count_matrix) >= 3:
                anomaly_score = self.anomaly_score()
                if anomaly_score > self.min_anomaly_score:
                    try:
                        data = log_atom.raw_data.decode(AminerConfig.ENCODING)
                    except UnicodeError:
                        data = repr(log_atom.raw_data)
                    if self.output_logline:
                        original_log_line_prefix = self.aminer_config.config_properties.get(
                            CONFIG_KEY_LOG_LINE_PREFIX, DEFAULT_LOG_LINE_PREFIX)
                        sorted_log_lines = [log_atom.parser_match.match_element.annotate_match('') + os.linesep + original_log_line_prefix +
                                            data]
                    else:
                        sorted_log_lines = [data]
                    affected_paths = []
                    affected_values = []
                    affected_counts = []
                    for path, count_dict in self.event_count_vector.items():
                        affected_paths.append(path)
                        affected_values.append(list(count_dict.keys()))
                        affected_counts.append(list(count_dict.values()))
                    analysis_component = {'AffectedLogAtomPaths': affected_paths, 'AffectedLogAtomValues': affected_values,
                                          'AffectedValueCounts': affected_counts, 'AnomalyScore': anomaly_score[0]}
                    event_data = {'AnalysisComponent': analysis_component}
                    for listener in self.anomaly_event_handlers:
                        listener.receive_event(f'Analysis.{self.__class__.__name__}', 'PCA anomaly detected', sorted_log_lines, event_data,
                                               log_atom, self)
            self.log_windows += 1

            # Add new values to matrix in learn mode
            if self.learn_mode is True:
                if len(self.event_count_matrix) >= self.num_windows:
                    del self.event_count_matrix[0]
                self.event_count_matrix.append(copy.deepcopy(self.event_count_vector))
                # PCA computation only possible when at least 3 vectors are present
                if len(self.event_count_matrix) >= 3:
                    self.repair_dict()
                    self.compute_pca()
                if self.stop_learning_timestamp is not None and self.stop_learning_no_anomaly_time is not None:
                    self.stop_learning_timestamp = time.time() + self.stop_learning_no_anomaly_time

            # Set window end time for next iteration
            self.start_time += self.window_size
            # Reset count vector for next time window
            self.reset_event_count_vector()

        if self.target_path_list is None or len(self.target_path_list) == 0:
            # Event is defined by the full path of log atom.
            constraint_path_flag = False
            for constraint_path in self.constraint_list:
                if parser_match.get_match_dictionary().get(constraint_path) is not None:
                    constraint_path_flag = True
                    break
            if not constraint_path_flag and self.constraint_list != []:
                return
            log_event = tuple(parser_match.get_match_dictionary().keys())
            if log_event in self.event_count_vector['']:
                self.event_count_vector[''][log_event] += 1
            else:
                self.event_count_vector[''][log_event] = 1
        else:
            # Event is defined by values in target_path_list
            all_values_none = True
            for path in self.target_path_list:
                match = parser_match.get_match_dictionary().get(path)
                if match is None:
                    continue
                matches = []
                if isinstance(match, list):
                    matches = match
                else:
                    matches.append(match)
                for match in matches:
                    if isinstance(match.match_object, bytes):
                        value = match.match_object.decode(AminerConfig.ENCODING)
                    else:
                        value = str(match.match_object)
                    if value is not None:
                        all_values_none = False
                    if path in self.event_count_vector:
                        if value in self.event_count_vector[path]:
                            self.event_count_vector[path][value] += 1
                        else:
                            self.event_count_vector[path][value] = 1
                    else:
                        self.event_count_vector[path] = {value: 1}
            if all_values_none is True:
                return
        self.log_success += 1

    def compute_pca(self):
        """Carry out PCA on current event count matrix."""
        # extract the features out of ecm into a list
        self.feature_list = []
        for events in self.event_count_matrix[0].values():
            for feature in events:
                self.feature_list.append(feature)

        # extract existing event_counts into array
        matrix = []
        for event_count in self.event_count_matrix:
            row = []
            for event in event_count.values():
                row += list(event.values())
            matrix.append(row)
        self.ecm = np.array(matrix)

        # Principal Component Analysis (PCA)
        normalized_ecm = (self.ecm - self.ecm.mean()) / self.ecm.std()
        covariance_matrix = np.cov(normalized_ecm.T)
        eigen_values, eigen_vectors = np.linalg.eigh(covariance_matrix)
        self.pca_ecm = normalized_ecm @ eigen_vectors
        self.eigen_vectors = eigen_vectors

        # number of components (n_comp): how many components should be used for reconstruction
        self.n_comp = self.get_n_comp(eigen_values)

        # PCA Inverse with only these components which describes the min_variance
        pca_inverse = self.pca_ecm[:, :self.n_comp] @ eigen_vectors[:self.n_comp, :]

        # Calculate Anomaly-Score (Reconstruction Error) for the whole dataset
        self.loss = np.sum((normalized_ecm - pca_inverse)**2, axis=1)

    def anomaly_score(self):
        """Calculate the anomalyscore for current event_count_vector."""
        # convert the event_count_vector into an array
        ecv = self.vector2array()
        # normalize the ecv with the mean and std of learned ecm
        normalized_ecv = (ecv - self.ecm.mean()) / self.ecm.std()
        # reshape array into a 1-dimensional array
        normalized_ecv = normalized_ecv.reshape(1, -1)
        # calculate the reduced pca for current log-sequence with given eigen_vectors
        pca_ecv = normalized_ecv @ self.eigen_vectors
        # calculate the pca_inverse with reduced number of components / do reconstruction
        pca_inverse_ecv = pca_ecv[:, :self.n_comp] @ self.eigen_vectors[:self.n_comp, :]
        # calculate the reconstruction error / anomaly score
        loss = np.sum((normalized_ecv - pca_inverse_ecv)**2, axis=1)
        # scale the reconstruction error with the min, max of ecm-loss
        loss = (loss - np.min(self.loss)) / (np.max(self.loss) - np.min(self.loss))
        return loss

    def vector2array(self):
        """Extract only the values which were learned before from current self.event_count_vector and return an array."""
        vector = []
        for event in self.event_count_vector.values():
            for feature, value in event.items():
                if feature in self.feature_list:
                    vector.append(value)
        return np.array(vector)

    def get_n_comp(self, eigen_values):
        """Return the number of components, which describe the variance threshold."""
        # Calculate the explained variance on each of components
        variance_explained = []
        for i in eigen_values[::-1]:
            variance_explained.append((i/sum(eigen_values))*100)
        # Calculate the cumulative explained variance (np.cumsum)
        cumulative_variance_explained = np.cumsum(variance_explained)
        for n, i in enumerate(cumulative_variance_explained):
            if i > (self.min_variance * 100):
                return n
        return None

    def repair_dict(self):
        """Check if any new values were added in current event_count_vector and repair self.event_count_matrix when necessary."""
        for ecv in self.event_count_matrix:
            for key, value in self.event_count_vector.items():
                if key not in ecv.keys():
                    for val in value:
                        ecv[key] = {val: 0}
                if not self.event_count_vector[key].keys() == ecv[key].keys():
                    for k in self.event_count_vector[key].keys():
                        if k not in ecv[key].keys():
                            ecv[key][k] = 0

    def reset_event_count_vector(self):
        """Reset event_count_vector by setting all count-values to 0."""
        for events in self.event_count_vector.values():
            for value in events:
                events[value] = 0

    def do_timer(self, trigger_time):
        """Check if current ruleset should be persisted."""
        if self.next_persist_time is None:
            return self.aminer_config.config_properties.get(KEY_PERSISTENCE_PERIOD, DEFAULT_PERSISTENCE_PERIOD)

        delta = self.next_persist_time - trigger_time
        if delta <= 0:
            self.do_persist()
            delta = self.aminer_config.config_properties.get(KEY_PERSISTENCE_PERIOD, DEFAULT_PERSISTENCE_PERIOD)
            self.next_persist_time = time.time() + delta
        return delta

    def do_persist(self):
        """Immediately write persistence data to storage."""
        if self.learn_mode is True:
            PersistenceUtil.store_json(self.persistence_file_name, list(self.event_count_matrix))

    def allowlist_event(self, event_type, event_data, allowlisting_data):
        """
        Allowlist an event generated by this source using the information emitted when generating the event.
        @return a message with information about allowlisting
        @throws Exception when allowlisting of this special event using given allowlisting_data was not possible.
        """
        if event_type != f'Analysis.{self.__class__.__name__}':
            msg = 'Event not from this source'
            logging.getLogger(AminerConfig.DEBUG_LOG_NAME).error(msg)
            raise Exception(msg)
        if allowlisting_data is not None:
            msg = 'Allowlisting data not understood by this detector'
            logging.getLogger(AminerConfig.DEBUG_LOG_NAME).error(msg)
            raise Exception(msg)
        if event_data not in self.constraint_list:
            self.constraint_list.append(event_data)
        return f'Allowlisted path {event_data}.'

    def blocklist_event(self, event_type, event_data, blocklisting_data):
        """
        Blocklist an event generated by this source using the information emitted when generating the event.
        @return a message with information about blocklisting
        @throws Exception when blocklisting of this special event using given blocklisting_data was not possible.
        """
        if event_type != f'Analysis.{self.__class__.__name__}':
            msg = 'Event not from this source'
            logging.getLogger(AminerConfig.DEBUG_LOG_NAME).error(msg)
            raise Exception(msg)
        if blocklisting_data is not None:
            msg = 'Blocklisting data not understood by this detector'
            logging.getLogger(AminerConfig.DEBUG_LOG_NAME).error(msg)
            raise Exception(msg)
        if event_data not in self.ignore_list:
            self.ignore_list.append(event_data)
        return f'Blocklisted path {event_data}.'

    def log_statistics(self, component_name):
        """
        Log statistics of an AtomHandler. Override this method for more sophisticated statistics output of the AtomHandler.
        @param component_name the name of the component which is printed in the log line.
        """
        if STAT_LEVEL == 1:
            logging.getLogger(STAT_LOG_NAME).info(
                f"'{component_name}' processed {self.log_success} out of {self.log_total} log atoms successfully in {self.log_windows} "
                f"time windows in the last 60 minutes.")
        elif STAT_LEVEL == 2:
            logging.getLogger(STAT_LOG_NAME).info(
                f"'{component_name}' processed {self.log_success} out of {self.log_total} log atoms successfully in {self.log_windows} "
                f"time windows in the last 60 minutes.")
        self.log_success = 0
        self.log_total = 0
        self.log_windows = 0



"""
This file contains interface definition useful implemented by classes in this directory and for use from code outside this directory.
All classes are defined in separate files, only the namespace references are added here to simplify the code.

This program is free software: you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.
This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
You should have received a copy of the GNU General Public License along with
this program. If not, see <http://www.gnu.org/licenses/>.

"""
import abc
import time
import logging
from aminer.AminerConfig import STAT_LOG_NAME, DEBUG_LOG_NAME, KEY_PERSISTENCE_PERIOD, DEFAULT_PERSISTENCE_PERIOD
from aminer import AminerConfig


class AtomizerFactory(metaclass=abc.ABCMeta):
    """
    This is the common interface of all factories to create atomizers for new data sources.
    These atomizers are integrated into the downstream processing pipeline.
    """

    @abc.abstractmethod
    def get_atomizer_for_resource(self, resource_name):
        """
        Get an atomizer for a given resource.
        @return a StreamAtomizer object
        """


class StreamAtomizer(metaclass=abc.ABCMeta):
    """
    This is the common interface of all binary stream atomizers.
    Atomizers in general should be good detecting and reporting malformed atoms but continue to function by attempting error correction or
    resynchronization with the stream after the bad atom. This type of atomizer also signals a stream source when the stream data cannot be
    handled at the moment to throttle reading  of the underlying stream.
    """

    @abc.abstractmethod
    def consume_data(self, stream_data, end_of_stream_flag=False):
        """
        Consume data from the underlying stream for atomizing. Data should only be consumed after splitting of an atom.
        The caller has to keep unconsumed data till the next invocation.
        @param stream_data the data offered to be consumed or zero length data when endOfStreamFlag is True (see below).
        @param end_of_stream_flag this flag is used to indicate, that the streamData offered is the last from the input stream.
        If the streamData does not form a complete atom, no rollover is expected or rollover would have honoured the atom boundaries,
        then the StreamAtomizer should treat that as an error. With rollover, consuming of the stream end data will signal the
        invoker to continue with data from next stream. When end of stream was reached but invoker has no streamData to send,
        it will invoke this method with zero-length data, which has to be consumed with a zero-length reply.
        @return the number of consumed bytes, 0 if the atomizer would need more data for a complete atom or -1 when no data was
        consumed at the moment but data might be consumed later on. The only situation where 0 is not an allowed return value
        is when end_of_stream_flag is set and stream_data not empty.
        """


class AtomHandlerInterface(metaclass=abc.ABCMeta):
    """This is the common interface of all handlers suitable for receiving log atoms."""

    output_event_handlers = None

    def __init__(self, mutable_default_args=None, learn_mode=None, stop_learning_time=None, stop_learning_no_anomaly_time=None,
                 stop_when_handled_flag=None, **kwargs):
        """Initialize the parameters of analysis components. See the classes of the analysis components for parameter descriptions."""
        allowed_kwargs = [
            "mutable_default_args", "aminer_config", "anomaly_event_handlers", "learn_mode", "persistence_id", "id_path_list",
            "stop_learning_time", "stop_learning_no_anomaly_time", "output_logline", "target_path_list", "constraint_list", "ignore_list",
            "allowlist_rules", "subhandler_list", "stop_when_handled_flag", "parsed_atom_handler_lookup_list",
            "default_parsed_atom_handler", "target_path", "parsed_atom_handler_dict", "allow_missing_values_flag",
            "tuple_transformation_function", "prob_thresh", "skip_repetitions", "max_hypotheses", "hypothesis_max_delta_time",
            "generation_probability", "generation_factor", "max_observations", "p0", "alpha", "candidates_size",
            "hypotheses_eval_delta_time", "delta_time_to_discard_hypothesis", "check_rules_flag", "window_size", "scoring_path_list",
            "num_windows", "confidence_factor", "empty_window_warnings", "early_exceeding_anomaly_output", "set_lower_limit",
            "set_upper_limit", "local_maximum_threshold", "seq_len", "allow_missing_id", "timeout", "allowed_id_tuples", "min_num_vals",
            "max_num_vals", "save_values", "track_time_for_tsa", "waiting_time", "num_sections_waiting_time", "histogram_definitions",
            "report_interval", "reset_after_report_flag", "bin_definition", "target_value_list", "timestamp_path", "min_bin_elements",
            "min_bin_time", "debug_mode", "stream", "separator", "missing_value_string", "num_log_lines_solidify_matrix",
            "time_output_threshold", "anomaly_threshold", "default_interval", "realert_interval", "combine_values", "min_allowed_time_diff",
            "target_label_list", "split_reports_flag", "event_type_detector", "num_init", "force_period_length", "set_period_length",
            "alpha_bt", "num_results_bt", "num_min_time_history", "num_max_time_history", "num_periods_tsa_ini", "time_period_length",
            "max_time_diff", "num_reduce_time_list", "min_anomaly_score", "min_variance", "parallel_check_count",
            "record_count_before_event", "use_path_match", "use_value_match", "min_rule_attributes", "max_rule_attributes", "ruleset",
            "exit_on_error_flag", "acf_pause_interval_percentage", "acf_auto_pause_interval", "acf_auto_pause_interval_num_min",
            "build_sum_over_values", "num_division_time_step", "acf_threshold", "round_time_interval_threshold",
            "min_log_lines_per_time_step", "num_update", "disc_div_thres", "num_steps_create_new_rules", "num_upd_until_validation",
            "num_end_learning_phase", "check_cor_thres", "check_cor_prob_thres", "check_cor_num_thres", "min_values_cors_thres",
            "new_vals_alarm_thres", "num_bt", "used_homogeneity_test", "alpha_chisquare_test", "max_dist_rule_distr", "used_presel_meth",
            "intersect_presel_meth", "percentage_random_cors", "match_disc_vals_sim_tresh", "exclude_due_distr_lower_limit",
            "match_disc_distr_threshold", "used_cor_meth", "used_validate_cor_meth", "validate_cor_cover_vals_thres",
            "validate_cor_distinct_thres", "used_gof_test", "gof_alpha", "s_gof_alpha", "s_gof_bt_alpha", "d_alpha", "d_bt_alpha",
            "div_thres", "sim_thres", "indicator_thres", "num_update_unq", "num_s_gof_values", "num_s_gof_bt", "num_d_bt",
            "num_pause_discrete", "num_pause_others", "test_gof_int", "num_stop_update", "silence_output_without_confidence",
            "silence_output_except_indicator", "num_var_type_hist_ref", "num_update_var_type_hist_ref", "num_var_type_considered_ind",
            "num_stat_stop_update", "num_updates_until_var_reduction", "var_reduction_thres", "num_skipped_ind_for_weights",
            "num_ind_for_weights", "used_multinomial_test", "use_empiric_distr", "used_range_test", "range_alpha", "range_threshold",
            "num_reinit_range", "range_limits_factor", "dw_alpha", "save_statistics", "idf", "norm", "add_normal", "check_empty_windows",
            "unique_path_list"
        ]
        self.log_success = 0
        self.log_total = 0
        self.persistence_id = None  # persistence_id is always needed.
        for argument, value in list(locals().items())[1:-1]:  # skip self parameter and kwargs
            if value is not None:
                setattr(self, argument, value)
        for argument, value in kwargs.items():  # skip self parameter and kwargs
            if argument not in allowed_kwargs:
                msg = f"Argument {argument} is unknown. Consider changing it or adding it to the allowed_kwargs list."
                logging.getLogger(DEBUG_LOG_NAME).error(msg)
                raise ValueError(msg)
            setattr(self, argument, value)

        if learn_mode is False and (stop_learning_time is not None or stop_learning_no_anomaly_time is not None):
            msg = "It is not possible to use the stop_learning_time or stop_learning_no_anomaly_time when the learn_mode is False."
            logging.getLogger(DEBUG_LOG_NAME).error(msg)
            raise ValueError(msg)
        if stop_learning_time is not None and stop_learning_no_anomaly_time is not None:
            msg = "stop_learning_time is mutually exclusive to stop_learning_no_anomaly_time. Only one of these attributes may be used."
            logging.getLogger(DEBUG_LOG_NAME).error(msg)
            raise ValueError(msg)
        if not isinstance(stop_learning_time, (type(None), int)):
            msg = "stop_learning_time has to be of the type int or None."
            logging.getLogger(DEBUG_LOG_NAME).error(msg)
            raise TypeError(msg)
        if not isinstance(stop_learning_no_anomaly_time, (type(None), int)):
            msg = "stop_learning_no_anomaly_time has to be of the type int or None."
            logging.getLogger(DEBUG_LOG_NAME).error(msg)
            raise TypeError(msg)

        self.stop_learning_timestamp = None
        if stop_learning_time is not None:
            self.stop_learning_timestamp = time.time() + stop_learning_time
        self.stop_learning_no_anomaly_time = stop_learning_no_anomaly_time
        if stop_learning_no_anomaly_time is not None:
            self.stop_learning_timestamp = time.time() + stop_learning_no_anomaly_time

        if hasattr(self, "aminer_config"):
            self.next_persist_time = time.time() + self.aminer_config.config_properties.get(
                KEY_PERSISTENCE_PERIOD, DEFAULT_PERSISTENCE_PERIOD)

        if mutable_default_args is not None:
            for argument in mutable_default_args:
                if hasattr(self, argument) and getattr(self, argument) is not None:
                    continue
                if argument.endswith("list"):
                    setattr(self, argument, [])
                elif argument.endswith("dict"):
                    setattr(self, argument, {})
                elif argument.endswith("set"):
                    setattr(self, argument, set())
                elif argument.endswith("tuple"):
                    setattr(self, argument, ())

        if hasattr(self, "subhandler_list"):
            if (not isinstance(self.subhandler_list, list)) or \
                    (not all(isinstance(handler, AtomHandlerInterface) for handler in self.subhandler_list)):
                msg = "Only subclasses of AtomHandlerInterface allowed in subhandler_list."
                logging.getLogger(DEBUG_LOG_NAME).error(msg)
                raise TypeError(msg)
            for handler_pos, handler_element in enumerate(self.subhandler_list):
                self.subhandler_list[handler_pos] = (handler_element, stop_when_handled_flag)
        if hasattr(self, "allowed_id_tuples"):
            if self.allowed_id_tuples is None:
                self.allowed_id_tuples = []
            else:
                self.allowed_id_tuples = [tuple(tuple_list) for tuple_list in self.allowed_id_tuples]

        if hasattr(self, "confidence_factor") and not 0 <= self.confidence_factor <= 1:
            logging.getLogger(DEBUG_LOG_NAME).warning('confidence_factor must be in the range [0,1]!')
            self.confidence_factor = 1

    @abc.abstractmethod
    def receive_atom(self, log_atom):
        """
        Receive a log atom from a source.
        @param log_atom binary raw atom data
        @return True if this handler was really able to handle and process the atom. Depending on this information, the caller
        may decide if it makes sense passing the atom also to other handlers or to retry later. This behaviour has to be documented
        at each source implementation sending LogAtoms.
        """

    def log_statistics(self, component_name):
        """
        Log statistics of an AtomHandler. Override this method for more sophisticated statistics output of the AtomHandler.
        @param component_name the name of the component which is printed in the log line.
        """
        if AminerConfig.STAT_LEVEL > 0:
            logging.getLogger(STAT_LOG_NAME).info(
                "'%s' processed %d out of %d log atoms successfully in the last 60 minutes.", component_name, self.log_success,
                self.log_total)
        self.log_success = 0
        self.log_total = 0


class LogDataResource(metaclass=abc.ABCMeta):
    """
    This is the superinterface of each logdata resource monitored by aminer.
    The interface is designed in a way, that instances of same subclass can be used both on aminer parent process side for keeping track of
    the resources and forwarding the file descriptors to the child, but also on child side for the same purpose. The only difference is,
    that on child side, the stream reading and read continuation features are used also. After creation on child side, this is the sole
    place for reading and closing the streams. An external process may use the file descriptor only to wait for input via select.
    """

    @abc.abstractmethod
    def __init__(self, log_resource_name, log_stream_fd, default_buffer_size=1 << 16, repositioning_data=None):
        """
        Create a new LogDataResource. Object creation must not touch the logStreamFd or read any data, unless repositioning_data  was given.
        In the later case, the stream has to support seek operation to reread data.
        @param log_resource_name the unique encoded name of this source as byte array.
        @param log_stream_fd the stream for reading the resource or -1 if not yet opened.
        @param repositioning_data if not None, attemt to position the the stream using the given data.
        """

    @abc.abstractmethod
    def open(self, reopen_flag=False):
        """
        Open the given resource.
        @param reopen_flag when True, attempt to reopen the same resource and check if it differs from the previously opened one.
        @raise Exception if valid logStreamFd was already provided, is still open and reopenFlag is False.
        @raise OSError when opening failed with unexpected error.
        @return True if the resource was really opened or False if opening was not yet possible but should be attempted again.
        """

    @abc.abstractmethod
    def get_resource_name(self):
        """Get the name of this log resource."""

    @abc.abstractmethod
    def get_file_descriptor(self):
        """Get the file descriptor of this open resource."""

    @abc.abstractmethod
    def fill_buffer(self):
        """
        Fill the buffer data of this resource. The repositioning information is not updated, update_position() has to be used.
        @return the number of bytes read or -1 on error or end.
        """

    @abc.abstractmethod
    def update_position(self, length):
        """Update the positioning information and discard the buffer data afterwards."""

    @abc.abstractmethod
    def get_repositioning_data(self):
        """Get the data for repositioning the stream. The returned structure has to be JSON serializable."""

    @abc.abstractmethod
    def close(self):
        """Close this logdata resource. Data access methods will not work any more afterwards."""



"""
This module defines a model element for parsing json strings

This program is free software: you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.
This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
You should have received a copy of the GNU General Public License along with
this program. If not, see <http://www.gnu.org/licenses/>.
"""
import logging
import orjson
from collections import deque

from typing import Any
from aminer.AminerConfig import DEBUG_LOG_NAME
from aminer.parsing.MatchContext import MatchContext
from aminer.parsing.MatchElement import MatchElement
from aminer.parsing.ModelElementInterface import ModelElementInterface


class JsonAccessObject:
    """
    The JsonAccessObject transforms a dictionary. It takes a dictionary "d" and
    flattens the dictionary to: key.another_key.somelist[0].foo = bar
    During the flatten()-process, it will create a self.collection dictionary with
    the format: collection[flattened-key]{levels[],value}
    """

    def __init__(self, d: dict):
        self.debug: bool = False
        self.levels: deque = deque()
        self.delimiter: str = '.'
        self.collection: dict = {}
        self.flatten(d)

    def join_levels(self):
        """joins levels using a specific delimiter"""
        ret = ""
        for i in self.levels:
            if not i.startswith("[") and len(ret) != 0:
                ret += self.delimiter
            ret += i
        return ret

    def create_collection_entry(self, index: str, levels: deque, value):
        """adds entry to the collection"""
        subentry = {}
        subentry['levels'] = levels.copy()
        subentry['value'] = value
        self.collection[index] = subentry

    def flatten(self, d: Any, islist=-1):
        """recursive function for flattening a dictionary"""
        if islist > -1:
            for k in d:
                if isinstance(k, dict):
                    # skipcq: FLK-E228
                    self.levels.append(f"[{islist}]")
                    islist = islist+1
                    self.flatten(k)
                    self.levels.pop()
                elif isinstance(k, list):
                    self.flatten(k, list)
                else:
                    if self.debug:
                        print(f"{ self.join_levels() }[{ islist }]: { k }")
                    # skipcq: PYL-C0209
                    self.create_collection_entry("%s[%d]" % (self.join_levels(), islist), self.levels, k)
                    islist = islist + 1
        else:
            for (k, v) in d.items():
                if isinstance(v, dict):
                    self.levels.append(k)
                    self.flatten(v)
                    if len(self.levels) != 0:
                        self.levels.pop()
                elif isinstance(v, list):
                    self.levels.append(k)
                    self.flatten(v, 0)
                    if len(self.levels) != 0:
                        self.levels.pop()
                else:
                    if len(self.levels) == 0:
                        if self.debug:
                            print(f"{ k } : { v }")
                        self.create_collection_entry(k, deque([k]), v)
                    else:
                        if islist > -1:
                            # skipcq: FLK-E228
                            self.levels.append(f"{k}[{ islist}]")
                            islist = islist+1
                        else:
                            self.levels.append(k)
                        if self.debug:
                            print(f"{ self.join_levels() } : { v }")
                        self.create_collection_entry(self.join_levels(), self.levels, v)
                        self.levels.pop()


class JsonStringModelElement(ModelElementInterface):
    """This class parses json-strings and matches the keys with a given key_parser_dict."""

    def __init__(self, element_id: str, key_parser_dict: dict, strict_mode: bool = False, ignore_null: bool = True):
        self.children: list = []

        self.strict_mode = strict_mode
        self.ignore_null = ignore_null

        if not isinstance(key_parser_dict, dict):
            msg = "key_parser_dict has to be of the type dict."
            logging.getLogger(DEBUG_LOG_NAME).error(msg)
            raise TypeError(msg)
        self.jao = JsonAccessObject(key_parser_dict)
        if not isinstance(element_id, str):
            msg = "element_id has to be of the type string."
            logging.getLogger(DEBUG_LOG_NAME).error(msg)
            raise TypeError(msg)
        if len(element_id) < 1:
            msg = "element_id must not be empty."
            logging.getLogger(DEBUG_LOG_NAME).error(msg)
            raise ValueError(msg)
        self.element_id = element_id
        self.fill_children()
        super().__init__(element_id, key_parser_dict=key_parser_dict, strict_mode=strict_mode,
                         ignore_null=ignore_null)

    def fill_children(self):
        """creates list of children from config-json"""
        for entry in self.jao.collection.values():
            self.children.append(entry['value'])

    def get_id(self):
        """Get the element ID."""
        return self.element_id

    def get_child_elements(self):  # skipcq: PYL-R0201
        """
        Get all possible child model elements of this element.
        @return None as there are no children of this element.
        """
        return self.children

    def get_match_element(self, path: str, match_context):
        """Just return a match including all data from the context."""
        current_path = f"{ path }/ { self.element_id }"
        logging.getLogger(DEBUG_LOG_NAME).info("JsonStringModelElement %s/%s", path, match_context.match_data.decode('utf-8'))
        matches = []
        try:
            jdict = orjson.loads(match_context.match_data)
            if self.strict_mode:
                jdictjao = JsonAccessObject(jdict)
                if len(jdictjao.collection) != len(self.jao.collection):
                    msg = "JsonStringModelElement-subparser-error: "
                    msg += "strict mode enabled and fields detected that do not exist in parser-config"
                    logging.getLogger(DEBUG_LOG_NAME).debug(msg)
                    return None
                try:
                    for (k, v) in self.jao.collection.items():
                        # empty string if value is null
                        parse_line = b""
                        if jdictjao.collection[k]['value'] is not None:
                            parse_line = str(jdictjao.collection[k]['value']).encode('utf-8')
                        else:
                            if self.ignore_null:
                                logging.getLogger(DEBUG_LOG_NAME).debug("JsonStringModelElement: ignore null at %s", k)
                                continue
                        child_match = v['value'].get_match_element(current_path, MatchContext(parse_line))
                        if child_match is None:
                            msg = "JsonStringModelElement-subparser-error: %s -> %s"
                            logging.getLogger(DEBUG_LOG_NAME).debug(msg, k, str(jdictjao.collection[k]['value']))
                            return None
                        matches += [child_match]
                except KeyError:
                    msg = "JsonStringModelElement-subparser-error: field \"%s\" not found but strict-enabled"
                    logging.getLogger(DEBUG_LOG_NAME).debug(msg, k)
                    return None
            else:
                for (k, v) in self.jao.collection.items():
                    tmp = jdict.copy()
                    try:
                        for level in v['levels']:
                            tmp = tmp[level]
                    except KeyError:
                        logging.getLogger(DEBUG_LOG_NAME).debug("JsonStringModelElement-subparser: %s not found", k)
                    parse_line = b""
                    # empty string if value is null
                    if tmp is not None:
                        parse_line = str(tmp).encode('utf-8')
                    else:
                        if self.ignore_null:
                            logging.getLogger(DEBUG_LOG_NAME).debug("JsonStringModelElement: ignore null at %s", k)
                            continue
                    child_match = v['value'].get_match_element(current_path, MatchContext(parse_line))
                    if child_match is None:
                        logging.getLogger(DEBUG_LOG_NAME).debug("JsonStringModelElement-subparser-error: %s -> %s", k, tmp)
                        return None
                    matches += [child_match]
        except orjson.JSONDecodeError as exception:
            msg = f"JsonStringModelElement { exception }: { match_context.match_data.decode('utf-8') }"
            logging.getLogger(DEBUG_LOG_NAME).error(msg)
            return None

        match_data = match_context.match_data
        if not match_data:
            return None
        match_context.update(match_data)
        return MatchElement(current_path, match_data, match_data, matches)



"""
This module defines an model element for integer number parsing.

This program is free software: you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.
This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
You should have received a copy of the GNU General Public License along with
this program. If not, see <http://www.gnu.org/licenses/>.
"""

import logging
from aminer.AminerConfig import DEBUG_LOG_NAME
from aminer.parsing.ModelElementInterface import ModelElementInterface
from aminer.parsing.MatchElement import MatchElement


class DecimalIntegerValueModelElement(ModelElementInterface):
    """
    This class defines a model to parse integer values with optional signum or padding.
    If both are present, it is signum has to be before the padding characters.
    """

    SIGN_TYPE_NONE = "none"
    SIGN_TYPE_OPTIONAL = "optional"
    SIGN_TYPE_MANDATORY = "mandatory"

    PAD_TYPE_NONE = "none"
    PAD_TYPE_ZERO = "zero"
    PAD_TYPE_BLANK = "blank"

    def __init__(self, element_id: str, value_sign_type: str = SIGN_TYPE_NONE, value_pad_type: str = PAD_TYPE_NONE):
        """
        Initialize the ModelElement.
        @param element_id an identifier for the ModelElement which is shown in the path.
        @param value_sign_type defines the possible start characters in the value. With the SIGN_TYPE_NONE only digits are allowed,
               with SIGN_TYPE_OPTIONAL digits and a minus sign are allowed and with SIGN_TYPE_MANDATORY the value must start with + or -.
        @param value_pad_type defines the padding values which can prefix the numerical value. With PAD_TYPE_NONE no padding is allowed,
               PAD_TYPE_ZERO allows zeros before the value and PAD_TYPE_BLANK allows spaces before the value.
        """
        super().__init__(element_id, value_sign_type=value_sign_type, value_pad_type=value_pad_type)

        if value_sign_type not in (DecimalIntegerValueModelElement.SIGN_TYPE_NONE, DecimalIntegerValueModelElement.SIGN_TYPE_OPTIONAL,
                                   DecimalIntegerValueModelElement.SIGN_TYPE_MANDATORY):
            msg = f"Invalid value_sign_type {value_sign_type}"
            logging.getLogger(DEBUG_LOG_NAME).error(msg)
            raise ValueError(msg)

        if value_pad_type not in (DecimalIntegerValueModelElement.PAD_TYPE_NONE, DecimalIntegerValueModelElement.PAD_TYPE_ZERO,
                                  DecimalIntegerValueModelElement.PAD_TYPE_BLANK):
            msg = f"Invalid value_pad_type {value_pad_type}"
            logging.getLogger(DEBUG_LOG_NAME).error(msg)
            raise ValueError(msg)
        self.digits = set(b"0123456789")

    def get_match_element(self, path: str, match_context):
        """
        Find the maximum number of bytes forming a integer number according to the parameters specified.
        @param path to be printed in the MatchElement.
        @param match_context the match_context to be analyzed.
        @return a match when at least one byte being a digit was found.
        """
        data = match_context.match_data

        if not data or (data[0] not in self.start_characters):
            return None
        match_len = 1

        if self.pad_characters == b"" and data.startswith(b"0") and not data.startswith(b"0.") and len(data) > 1 and \
                data[1] in self.digits:
            return None

        for test_byte in data[match_len:]:
            if test_byte not in self.pad_characters:
                break
            match_len += 1
        num_start_pos = match_len
        for test_byte in data[match_len:]:
            if test_byte not in self.digits:
                break
            match_len += 1

        if match_len == 1:  # skipcq: PTC-W0048
            if data[0] not in self.digits:
                return None
        elif num_start_pos == match_len and match_len == 1:  # only return None if match_len is 1 to allow 00 with zero padding.
            return None

        match_string = data[:match_len]
        try:
            if self.pad_characters == b" " and match_string[0] in b"+-":
                match_value = int(match_string.replace(b" ", b"", 1))
            else:
                match_value = int(match_string)
        except ValueError:
            return None
        match_context.update(match_string)
        return MatchElement(f"{path}/{self.element_id}", match_string, match_value, None)
